"""
Common functionality for converting metadata
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Union, Tuple, BinaryIO, Any, Optional
import hashlib
import os
import warnings
import struct
import mmap


import numpy

try:
    import h5py
except ImportError:
    h5py = None


###########
# general file type checks

def is_file_like(the_input: Any) -> bool:
    """
    Verify whether the provided input appear to provide a "file-like object". This
    term is used ubiquitously, but not all usages are identical. In this case, we
    mean that there exist callable attributes `read`, `write`, `seek`, and `tell`.

    Note that this does not check the mode (binary/string or read/write/append),
    as it is not clear that there is any generally accessible way to do so.

    Parameters
    ----------
    the_input

    Returns
    -------
    bool
    """

    out = True
    for attribute in ['read', 'write', 'seek', 'tell']:
        value = getattr(the_input, attribute, None)
        out &= callable(value)
    return out


def is_real_file(the_input: BinaryIO) -> bool:
    """
    Determine if the file-like object is associated with an actual file.
    This is mainly to consider suitability for establishment of a numpy.memmap.

    Parameters
    ----------
    the_input : BinaryIO

    Returns
    -------
    bool
    """

    if not hasattr(the_input, 'fileno'):
        return False
    # noinspection PyBroadException
    try:
        fileno = the_input.fileno()
        return isinstance(fileno, int) and (fileno >= 0)
    except Exception:
        return False


def _fetch_initial_bytes(file_name: Union[str, BinaryIO], size: int) -> Optional[bytes]:
    header = b''
    if is_file_like(file_name):
        current_location = file_name.tell()
        file_name.seek(0, os.SEEK_SET)
        header = file_name.read(size)
        file_name.seek(current_location, os.SEEK_SET)
    elif isinstance(file_name, str):
        if not os.path.isfile(file_name):
            return None

        with open(file_name, 'rb') as fi:
            header = fi.read(size)

    if len(header) != size:
        return None
    return header


def is_nitf(
        file_name: Union[str, BinaryIO],
        return_version=False) -> Union[bool, Tuple[bool, Optional[str]]]:
    """
    Test whether the given input is a NITF 2.0 or 2.1 file.

    Parameters
    ----------
    file_name : str|BinaryIO
    return_version : bool


    Returns
    -------
    is_nitf_file: bool
        Is the file a NITF file, based solely on checking initial bytes.
    nitf_version: None|str
        Only returned is `return_version=True`. Will be `None` in the event that
        `is_nitf_file=False`.
    """

    header = _fetch_initial_bytes(file_name, 9)
    if header is None:
        if return_version:
            return False, None
        else:
            return False

    ihead = header[:4]
    vers = header[4:]
    if ihead == b'NITF':
        try:
            vers = vers.decode('utf-8')
            return (True, vers) if return_version else True
        except ValueError:
            pass

    return (False, None) if return_version else False


def is_tiff(
        file_name: Union[str, BinaryIO],
        return_details=False) -> Union[bool, Tuple[bool, Optional[str], Optional[int]]]:
    """
    Test whether the given input is a tiff or big_tiff file.

    Parameters
    ----------
    file_name : str|BinaryIO
    return_details : bool
        Return the tiff details of endianness and magic number?

    Returns
    -------
    is_tiff_file : bool
    endianness : None|str
        Only returned if `return_details` is `True`. One of `['>', '<']`.
    magic_number : None|int
        Only returned if `return_details` is `True`. One of `[42, 43]`.
    """

    header = _fetch_initial_bytes(file_name, 4)
    if header is None:
        return (False, None, None) if return_details else False

    try:
        endian_part = header[:2].decode('utf-8')
    except ValueError:
        return (False, None, None) if return_details else False

    if endian_part not in ['II', 'MM']:
        return (False, None, None) if return_details else False

    if endian_part == 'II':
        endianness = '<'
    else:
        endianness = '>'
    magic_number = struct.unpack('{}h'.format(endianness), header[2:])[0]

    if magic_number in [42, 43]:
        # NB: 42 is regular tiff, while 43 is big tiff
        return (True, magic_number, endianness) if return_details else True
    return (False, None, None) if return_details else False


def is_hdf5(file_name: Union[str, BinaryIO]) -> bool:
    """
    Test whether the given input is a hdf5 file.

    Parameters
    ----------
    file_name : str|BinaryIO

    Returns
    -------
    bool
    """

    header = _fetch_initial_bytes(file_name, 4)
    if header is None:
        return False

    out = (header == b'\x89HDF')
    if out and h5py is None:
        warnings.warn('The h5py library was not successfully imported, and no hdf5 files can be read')
    return out


###########

def parse_timestring(str_in: str, precision: str = 'us') -> numpy.datetime64:
    """
    Parse (naively) a timestring to numpy.datetime64 of the given precision.

    Parameters
    ----------
    str_in : str
    precision : str
        See numpy.datetime64 for precision options.

    Returns
    -------
    numpy.datetime64
    """

    if str_in.strip()[-1] == 'Z':
        return numpy.datetime64(str_in[:-1], precision)
    return numpy.datetime64(str_in, precision)


def get_seconds(
        dt1: numpy.datetime64,
        dt2: numpy.datetime64,
        precision: str = 'us') -> float:
    """
    The number of seconds between two numpy.datetime64 elements.

    Parameters
    ----------
    dt1 : numpy.datetime64
    dt2 : numpy.datetime64
    precision : str
        one of 's', 'ms', 'us', or 'ns'.

    Returns
    -------
    float
        the number of seconds between dt2 and dt1 (i.e. dt1 - dt2).
    """

    if precision == 's':
        scale = 1
    elif precision == 'ms':
        scale = 1e-3
    elif precision == 'us':
        scale = 1e-6
    elif precision == 'ns':
        scale = 1e-9
    else:
        raise ValueError('unrecognized precision {}'.format(precision))

    dtype = 'datetime64[{}]'.format(precision)
    tdt1 = dt1.astype(dtype)
    tdt2 = dt2.astype(dtype)
    return float((tdt1.astype('int64') - tdt2.astype('int64'))*scale)


def calculate_md5(the_path: str, chunk_size: int = 1024*1024) -> str:
    """
    Calculate the md5 checksum of a given file defined by a path.

    Parameters
    ----------
    the_path : str
        The path to the file
    chunk_size : int
        The chunk size for processing

    Returns
    -------
    str
        The 32 character MD5 hex digest of the given file
    """

    md5_hash = hashlib.md5()
    with open(the_path, 'rb') as fi:
        for chunk in iter(lambda: fi.read(chunk_size), b''):
            md5_hash.update(chunk)
    return md5_hash.hexdigest()


#######
# Flexible memmap object for extracting compressed image data from mid-file

class MemMap(object):
    """
    Spoofing necessary memory map functionality to permit READ ONLY opening of a
    file containing compressed image data somewhere mid-file for use in the PIL
    interface. This is just a thin wrapper around the built-in Python memmap
    class which accommodates arbitrary offset (versus limited to allocation
    granularity).

    **The bare minimum of functionality is implemented to permit the intended use.**
    """

    __slots__ = ('_mem_map', '_file_obj', '_offset_shift')

    def __init__(self, file_obj, length, offset):
        """

        Parameters
        ----------
        file_obj : str|BinaryIO
        length : int
        offset : int
        """

        # length and offset validation
        length = int(length)
        offset = int(offset)
        if length < 0 or offset < 0:
            raise ValueError(
                'length ({}) and offset ({}) must be non-negative integers'.format(length, offset))
        # determine offset and length accommodating allocation block size limitation
        self._offset_shift = (offset % mmap.ALLOCATIONGRANULARITY)
        offset = offset - self._offset_shift
        length = length + self._offset_shift
        # establish the mem map
        if isinstance(file_obj, str):
            self._file_obj = open(file_obj, 'rb')
        else:
            self._file_obj = file_obj
        self._mem_map = mmap.mmap(self._file_obj.fileno(), length, access=mmap.ACCESS_READ, offset=offset)

    def read(self, n):
        return self._mem_map.read(n)

    def tell(self):
        return self._mem_map.tell() - self._offset_shift

    def seek(self, pos, whence=0):
        whence = int(whence)
        pos = int(pos)
        if whence == 0:
            self._mem_map.seek(pos+self._offset_shift, 0)
        else:
            self._mem_map.seek(pos, whence)

    @property
    def closed(self):
        return self._file_obj.closed

    def close(self):
        self._file_obj.close()




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class RSMIDAType(TREElement):
    def __init__(self, value):
        super(RSMIDAType, self).__init__()
        self.add_field('IID', 's', 80, value)
        self.add_field('EDITION', 's', 40, value)
        self.add_field('ISID', 's', 40, value)
        self.add_field('SID', 's', 40, value)
        self.add_field('STID', 's', 40, value)
        self.add_field('YEAR', 's', 4, value)
        self.add_field('MONTH', 's', 2, value)
        self.add_field('DAY', 's', 2, value)
        self.add_field('HOUR', 's', 2, value)
        self.add_field('MINUTE', 's', 2, value)
        self.add_field('SECOND', 's', 9, value)
        self.add_field('NRG', 's', 8, value)
        self.add_field('NCG', 's', 8, value)
        self.add_field('TRG', 's', 21, value)
        self.add_field('TCG', 's', 21, value)
        self.add_field('GRNDD', 's', 1, value)
        self.add_field('XUOR', 's', 21, value)
        self.add_field('YUOR', 's', 21, value)
        self.add_field('ZUOR', 's', 21, value)
        self.add_field('XUXR', 's', 21, value)
        self.add_field('XUYR', 's', 21, value)
        self.add_field('XUZR', 's', 21, value)
        self.add_field('YUXR', 's', 21, value)
        self.add_field('YUYR', 's', 21, value)
        self.add_field('YUZR', 's', 21, value)
        self.add_field('ZUXR', 's', 21, value)
        self.add_field('ZUYR', 's', 21, value)
        self.add_field('ZUZR', 's', 21, value)
        self.add_field('V1X', 's', 21, value)
        self.add_field('V1Y', 's', 21, value)
        self.add_field('V1Z', 's', 21, value)
        self.add_field('V2X', 's', 21, value)
        self.add_field('V2Y', 's', 21, value)
        self.add_field('V2Z', 's', 21, value)
        self.add_field('V3X', 's', 21, value)
        self.add_field('V3Y', 's', 21, value)
        self.add_field('V3Z', 's', 21, value)
        self.add_field('V4X', 's', 21, value)
        self.add_field('V4Y', 's', 21, value)
        self.add_field('V4Z', 's', 21, value)
        self.add_field('V5X', 's', 21, value)
        self.add_field('V5Y', 's', 21, value)
        self.add_field('V5Z', 's', 21, value)
        self.add_field('V6X', 's', 21, value)
        self.add_field('V6Y', 's', 21, value)
        self.add_field('V6Z', 's', 21, value)
        self.add_field('V7X', 's', 21, value)
        self.add_field('V7Y', 's', 21, value)
        self.add_field('V7Z', 's', 21, value)
        self.add_field('V8X', 's', 21, value)
        self.add_field('V8Y', 's', 21, value)
        self.add_field('V8Z', 's', 21, value)
        self.add_field('GRPX', 's', 21, value)
        self.add_field('GRPY', 's', 21, value)
        self.add_field('GRPZ', 's', 21, value)
        self.add_field('FULLR', 's', 8, value)
        self.add_field('FULLC', 's', 8, value)
        self.add_field('MINR', 's', 8, value)
        self.add_field('MAXR', 's', 8, value)
        self.add_field('MINC', 's', 8, value)
        self.add_field('MAXC', 's', 8, value)
        self.add_field('IE0', 's', 21, value)
        self.add_field('IER', 's', 21, value)
        self.add_field('IEC', 's', 21, value)
        self.add_field('IERR', 's', 21, value)
        self.add_field('IERC', 's', 21, value)
        self.add_field('IECC', 's', 21, value)
        self.add_field('IA0', 's', 21, value)
        self.add_field('IAR', 's', 21, value)
        self.add_field('IAC', 's', 21, value)
        self.add_field('IARR', 's', 21, value)
        self.add_field('IARC', 's', 21, value)
        self.add_field('IACC', 's', 21, value)
        self.add_field('SPX', 's', 21, value)
        self.add_field('SVX', 's', 21, value)
        self.add_field('SAX', 's', 21, value)
        self.add_field('SPY', 's', 21, value)
        self.add_field('SVY', 's', 21, value)
        self.add_field('SAY', 's', 21, value)
        self.add_field('SPZ', 's', 21, value)
        self.add_field('SVZ', 's', 21, value)
        self.add_field('SAZ', 's', 21, value)


class RSMIDA(TREExtension):
    _tag_value = 'RSMIDA'
    _data_type = RSMIDAType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class ACFTBType(TREElement):
    def __init__(self, value):
        super(ACFTBType, self).__init__()
        self.add_field('AC_MSN_ID', 's', 20, value)
        self.add_field('AC_TAIL_NO', 's', 10, value)
        self.add_field('AC_TO', 's', 12, value)
        self.add_field('SENSOR_ID_TYPE', 's', 4, value)
        self.add_field('SENSOR_ID', 's', 6, value)
        self.add_field('SCENE_SOURCE', 's', 1, value)
        self.add_field('SCNUM', 's', 6, value)
        self.add_field('PDATE', 's', 8, value)
        self.add_field('IMHOSTNO', 's', 6, value)
        self.add_field('IMREQID', 's', 5, value)
        self.add_field('MPLAN', 's', 3, value)
        self.add_field('ENTLOC', 's', 25, value)
        self.add_field('LOC_ACCY', 's', 6, value)
        self.add_field('ENTELV', 's', 6, value)
        self.add_field('ELV_UNIT', 's', 1, value)
        self.add_field('EXITLOC', 's', 25, value)
        self.add_field('EXITELV', 's', 6, value)
        self.add_field('TMAP', 's', 7, value)
        self.add_field('ROW_SPACING', 's', 7, value)
        self.add_field('ROW_SPACING_UNITS', 's', 1, value)
        self.add_field('COL_SPACING', 's', 7, value)
        self.add_field('COL_SPACING_UNITS', 's', 1, value)
        self.add_field('FOCAL_LENGTH', 's', 6, value)
        self.add_field('SENSERIAL', 's', 6, value)
        self.add_field('ABSWVER', 's', 7, value)
        self.add_field('CAL_DATE', 's', 8, value)
        self.add_field('PATCH_TOT', 's', 4, value)
        self.add_field('MTI_TOT', 's', 3, value)


class ACFTB(TREExtension):
    _tag_value = 'ACFTB'
    _data_type = ACFTBType



"""
This package contains the elements for interpreting product data.
"""

__classification__ = 'UNCLASSIFIED'


def open(*args, **kwargs):
    from .converter import open_product
    return open_product(*args, **kwargs)



"""
Functionality for reading a GFF file into a SICD model.

Note: This has been tested on files of version 1.8 and 2.5, but hopefully works for others.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
import os
import struct
from typing import Tuple, Union, BinaryIO, Optional
from datetime import datetime
from tempfile import mkstemp
import zlib

import numpy
from scipy.constants import speed_of_light

from sarpy.io.general.base import SarpyIOError
from sarpy.io.general.format_function import ComplexFormatFunction
from sarpy.io.general.data_segment import DataSegment, NumpyMemmapSegment
from sarpy.io.general.utils import is_file_like, MemMap
from sarpy.geometry.geocoords import geodetic_to_ecf, wgs_84_norm, ned_to_ecf

from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.CollectionInfo import CollectionInfoType, \
    RadarModeType
from sarpy.io.complex.sicd_elements.ImageCreation import ImageCreationType
from sarpy.io.complex.sicd_elements.ImageData import ImageDataType
from sarpy.io.complex.sicd_elements.GeoData import GeoDataType, SCPType
from sarpy.io.complex.sicd_elements.Grid import GridType, DirParamType, \
    WgtTypeType
from sarpy.io.complex.sicd_elements.SCPCOA import SCPCOAType
from sarpy.io.complex.sicd_elements.Timeline import TimelineType, IPPSetType
from sarpy.io.complex.sicd_elements.RadarCollection import RadarCollectionType, \
    WaveformParametersType, ChanParametersType
from sarpy.io.complex.sicd_elements.ImageFormation import ImageFormationType, \
    RcvChanProcType
from sarpy.io.complex.sicd_elements.Radiometric import RadiometricType, \
    NoiseLevelType_

try:
    import PIL
except ImportError:
    PIL = None

logger = logging.getLogger(__name__)

_requires_array_text = 'Requires numpy.ndarray, got `{}`'
_requires_3darray_text = 'Requires a three-dimensional numpy.ndarray\n\t' \
                         '(with band in the last dimension), got shape {}'


####################
# utility functions

def _get_string(bytes_in):
    bytes_in = bytes_in.replace(b'\x00', b'')
    return bytes_in.decode('utf-8')


def _rescale_float(int_in, scale):
    return float(int_in)/scale


####################
# version 1 specific header parsing

class _GFFHeader_1_6(object):
    """
    Interpreter for the GFF version 1.6 header
    """

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        self.file_object = fi
        self.estr = estr

        self.version = '1.6'
        fi.seek(12, os.SEEK_SET)
        # starting at line 3 of def
        self.header_length = struct.unpack(estr+'I', fi.read(4))[0]
        if self.header_length < 952:
            raise ValueError(
                'The provided header is apparently too short to be a version 1.6 GFF header')

        fi.read(2)  # redundant
        self.creator = _get_string(fi.read(24))
        self.date_time = struct.unpack(estr+'6H', fi.read(6*2))  # year,month, day, hour, minute, second
        fi.read(2)  # endian, already parsed
        self.bytes_per_pixel, self.frame_count, self.image_type, \
            self.row_major, self.range_count, self.azimuth_count = \
            struct.unpack(estr+'6I', fi.read(6*4))
        self.scale_exponent, self.scale_mantissa, self.offset_exponent, self.offset_mantissa = \
            struct.unpack(estr+'4i', fi.read(4*4))
        # at line 17 of def

        fi.read(2)  # redundant
        self.comment = _get_string(fi.read(166))
        self.image_plane = struct.unpack(estr+'I', fi.read(4))[0]
        range_pixel_size, azimuth_pixel_size, azimuth_overlap = struct.unpack(estr+'3I', fi.read(3*4))
        self.range_pixel_size = _rescale_float(range_pixel_size, 1 << 16)
        self.azimuth_pixel_size = _rescale_float(azimuth_pixel_size, 1 << 16)
        self.azimuth_overlap = _rescale_float(azimuth_overlap, 1 << 16)

        srp_lat, srp_lon, srp_alt, rfoa, x_to_srp = struct.unpack(estr+'5i', fi.read(5*4))
        self.srp_lat = _rescale_float(srp_lat, 1 << 23)
        self.srp_lon = _rescale_float(srp_lon, 1 << 23)
        self.srp_alt = _rescale_float(srp_alt, 1 << 16)
        self.rfoa = _rescale_float(rfoa, 1 << 23)
        self.x_to_srp = _rescale_float(x_to_srp, 1 << 16)

        fi.read(2)
        self.phase_name = _get_string(fi.read(128))
        fi.read(2)
        self.image_name = _get_string(fi.read(128))
        # at line 32 of def

        self.look_count, self.param_ref_ap, self.param_ref_pos = \
            struct.unpack(estr+'3I', fi.read(3*4))

        graze_angle, squint, gta, range_beam_ctr, flight_time = \
            struct.unpack(estr + 'I2i2I', fi.read(5*4))
        self.graze_angle = _rescale_float(graze_angle, 1 << 23)
        self.squint = _rescale_float(squint, 1 << 23)
        self.gta = _rescale_float(gta, 1 << 23)
        self.range_beam_ctr = _rescale_float(range_beam_ctr, 1 << 8)
        self.flight_time = _rescale_float(flight_time, 1000)

        self.range_chirp_rate, x_to_start, self.mo_comp_mode, v_x = \
            struct.unpack(estr+'fi2I', fi.read(4*4))
        self.x_to_start = _rescale_float(x_to_start, 1 << 16)
        self.v_x = _rescale_float(v_x, 1 << 16)
        # at line 44 of def

        apc_lat, apc_lon, apc_alt = struct.unpack(estr+'3i', fi.read(3*4))
        self.apc_lat = _rescale_float(apc_lat, 1 << 23)
        self.apc_lon = _rescale_float(apc_lon, 1 << 23)
        self.apc_alt = _rescale_float(apc_alt, 1 << 16)

        cal_parm, self.logical_block_address = struct.unpack(estr+'2I', fi.read(2*4))
        self.cal_parm = _rescale_float(cal_parm, 1 << 24)
        az_resolution, range_resolution = struct.unpack(estr+'2I', fi.read(2*4))
        self.az_resolution = _rescale_float(az_resolution, 1 << 16)
        self.range_resolution = _rescale_float(range_resolution, 1 << 16)

        des_sigma_n, des_graze, des_squint, des_range, scene_track_angle = \
            struct.unpack(estr+'iIiIi', fi.read(5*4))
        self.des_sigma_n = _rescale_float(des_sigma_n, 1 << 23)
        self.des_graze = _rescale_float(des_graze, 1 << 23)
        self.des_squint = _rescale_float(des_squint, 1 << 23)
        self.des_range = _rescale_float(des_range, 1 << 8)
        self.scene_track_angle = _rescale_float(scene_track_angle, 1 << 23)
        # at line 56 of def

        self.user_param = fi.read(48)  # leave uninterpreted

        self.coarse_snr, self.coarse_azimuth_sub, self.coarse_range_sub, \
            self.max_azimuth_shift, self.max_range_shift, \
            self.coarse_delta_azimuth, self.coarse_delta_range = \
            struct.unpack(estr+'7i', fi.read(7*4))

        self.tot_procs, self.tpt_box_cmode, self.snr_thresh, self.range_size, \
            self.map_box_size, self.box_size, self.box_spc, self.tot_tpts, \
            self.good_tpts, self.range_seed, self.range_shift, self.azimuth_shift = \
            struct.unpack(estr+'12i', fi.read(12*4))
        # at line 76 of def

        self.sum_x_ramp, self.sum_y_ramp = struct.unpack(estr+'2i', fi.read(2*4))
        self.cy9k_tape_block, self.nominal_center_frequency = struct.unpack(estr+'If', fi.read(2*4))
        self.image_flags, self.line_number, self.patch_number = struct.unpack(estr+'3I', fi.read(3*4))
        self.lambda0, self.srange_pix_space = struct.unpack(estr+'2f', fi.read(2*4))
        self.dopp_pix_space, self.dopp_offset, self.dopp_range_scale, self.mux_time_delay = \
            struct.unpack(estr+'4f', fi.read(4*4))
        # at line 89 of def

        self.apc_ecef = struct.unpack(estr+'3d', fi.read(3*8))
        self.vel_ecef = struct.unpack(estr+'3f', fi.read(3*4))
        self.phase_cal = struct.unpack(estr+'f', fi.read(4))[0]
        self.srp_ecef = struct.unpack(estr+'3d', fi.read(3*8))
        self.res5 = fi.read(64)  # leave uninterpreted


class _Radar_1_8(object):
    """
    The radar details, for version 1.8
    """

    def __init__(self, the_bytes, estr):
        """

        Parameters
        ----------
        the_bytes : bytes
            This will be required to have length 76
        estr : str
            The endianness format string
        """

        if not (isinstance(the_bytes, bytes) and len(the_bytes) == 76):
            raise ValueError('Incorrect length input')

        self.platform = _get_string(the_bytes[:24])
        self.proc_id = _get_string(the_bytes[24:36])
        self.radar_model = _get_string(the_bytes[36:48])
        self.radar_id = struct.unpack(estr+'I', the_bytes[48:52])[0]
        self.swid = _get_string(the_bytes[52:76])


class _GFFHeader_1_8(object):
    """
    Interpreter for the GFF version 1.8 header
    """

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        self.file_object = fi
        self.estr = estr

        self.version = '1.8'
        fi.seek(12, os.SEEK_SET)
        # starting at line 3 of def
        self.header_length = struct.unpack(estr+'I', fi.read(4))[0]
        if self.header_length < 2040:
            raise ValueError(
                'The provided header is apparently too short to be a version 1.8 GFF header')

        fi.read(2)  # redundant
        self.creator = _get_string(fi.read(24))
        self.date_time = struct.unpack(estr+'6H', fi.read(6*2))  # year, month, day, hour, minute, second
        fi.read(2)  # endian, already parsed
        self.bytes_per_pixel = int(struct.unpack(estr+'f', fi.read(4))[0])
        self.frame_count, self.image_type, self.row_major, self.range_count, \
            self.azimuth_count = struct.unpack(estr+'5I', fi.read(5*4))
        self.scale_exponent, self.scale_mantissa, self.offset_exponent, self.offset_mantissa = \
            struct.unpack(estr+'4i', fi.read(4*4))
        # at line 17 of def

        self.res1 = fi.read(32)  # leave uninterpreted

        fi.read(2)  # redundant
        self.comment = _get_string(fi.read(166))
        self.image_plane = struct.unpack(estr+'I', fi.read(4))[0]
        range_pixel_size, azimuth_pixel_size, azimuth_overlap = struct.unpack(estr+'3I', fi.read(3*4))
        self.range_pixel_size = _rescale_float(range_pixel_size, 1 << 16)
        self.azimuth_pixel_size = _rescale_float(azimuth_pixel_size, 1 << 16)
        self.azimuth_overlap = _rescale_float(azimuth_overlap, 1 << 16)

        srp_lat, srp_lon, srp_alt, rfoa, x_to_srp = struct.unpack(estr+'5i', fi.read(5*4))
        self.srp_lat = _rescale_float(srp_lat, 1 << 23)
        self.srp_lon = _rescale_float(srp_lon, 1 << 23)
        self.srp_alt = _rescale_float(srp_alt, 1 << 16)
        self.rfoa = _rescale_float(rfoa, 1 << 23)
        self.x_to_srp = _rescale_float(x_to_srp, 1 << 16)

        self.res2 = fi.read(32)  # leave uninterpreted

        fi.read(2)
        self.phase_name = _get_string(fi.read(128))
        fi.read(2)
        self.image_name = _get_string(fi.read(128))
        # at line 34 of def

        self.look_count, self.param_ref_ap, self.param_ref_pos = \
            struct.unpack(estr + '3I', fi.read(3*4))

        graze_angle, squint, gta, range_beam_ctr, flight_time = \
            struct.unpack(estr + 'I2i2I', fi.read(5*4))
        self.graze_angle = _rescale_float(graze_angle, 1 << 23)
        self.squint = _rescale_float(squint, 1 << 23)
        self.gta = _rescale_float(gta, 1 << 23)
        self.range_beam_ctr = _rescale_float(range_beam_ctr, 1 << 8)
        self.flight_time = _rescale_float(flight_time, 1000)

        self.range_chirp_rate, x_to_start, self.mo_comp_mode, v_x = \
            struct.unpack(estr + 'fi2I', fi.read(4*4))
        self.x_to_start = _rescale_float(x_to_start, 1 << 16)
        self.v_x = _rescale_float(v_x, 1 << 16)
        # at line 46 of def

        apc_lat, apc_lon, apc_alt = struct.unpack(estr + '3i', fi.read(3*4))
        self.apc_lat = _rescale_float(apc_lat, 1 << 23)
        self.apc_lon = _rescale_float(apc_lon, 1 << 23)
        self.apc_alt = _rescale_float(apc_alt, 1 << 16)

        cal_parm, self.logical_block_address = struct.unpack(estr + '2I', fi.read(2*4))
        self.cal_parm = _rescale_float(cal_parm, 1 << 24)
        az_resolution, range_resolution = struct.unpack(estr + '2I', fi.read(2*4))
        self.az_resolution = _rescale_float(az_resolution, 1 << 16)
        self.range_resolution = _rescale_float(range_resolution, 1 << 16)

        des_sigma_n, des_graze, des_squint, des_range, scene_track_angle = \
            struct.unpack(estr + 'iIiIi', fi.read(5*4))
        self.des_sigma_n = _rescale_float(des_sigma_n, 1 << 23)
        self.des_graze = _rescale_float(des_graze, 1 << 23)
        self.des_squint = _rescale_float(des_squint, 1 << 23)
        self.des_range = _rescale_float(des_range, 1 << 8)
        self.scene_track_angle = _rescale_float(scene_track_angle, 1 << 23)
        # at line 58 of def

        self.user_param = fi.read(48)  # leave uninterpreted

        self.coarse_snr, self.coarse_azimuth_sub, self.coarse_range_sub, \
        self.max_azimuth_shift, self.max_range_shift, \
        self.coarse_delta_azimuth, self.coarse_delta_range = \
            struct.unpack(estr + '7i', fi.read(7*4))

        self.tot_procs, self.tpt_box_cmode, self.snr_thresh, self.range_size, \
        self.map_box_size, self.box_size, self.box_spc, self.tot_tpts, \
        self.good_tpts, self.range_seed, self.range_shift, self.azimuth_shift = \
            struct.unpack(estr + '12i', fi.read(12*4))
        # at line 78 of def

        self.sum_x_ramp, self.sum_y_ramp = struct.unpack(estr + '2i', fi.read(2*4))
        self.cy9k_tape_block, self.nominal_center_frequency = struct.unpack(estr + 'If', fi.read(2*4))
        self.image_flags, self.line_number, self.patch_number = struct.unpack(estr + '3I', fi.read(3*4))
        self.lambda0, self.srange_pix_space = struct.unpack(estr + '2f', fi.read(2*4))
        self.dopp_pix_space, self.dopp_offset, self.dopp_range_scale, self.mux_time_delay = \
            struct.unpack(estr + '4f', fi.read(4*4))
        # at line 91 of def

        self.apc_ecef = struct.unpack(estr+'3d', fi.read(3*8))
        self.vel_ecef = struct.unpack(estr+'3f', fi.read(3*4))
        self.phase_cal = struct.unpack(estr+'f', fi.read(4))[0]
        self.srp_ecef = struct.unpack(estr+'3d', fi.read(3*8))

        self.res5 = fi.read(64)  # leave uninterpreted
        # at line 102

        self.header_length1 = struct.unpack(estr+'I', fi.read(4))[0]
        self.image_date = struct.unpack(estr+'6H', fi.read(6*2))  # year,month, day, hour, minute, second
        self.comp_file_name = _get_string(fi.read(128))
        self.ref_file_name = _get_string(fi.read(128))

        self.IE = _Radar_1_8(fi.read(76), estr)
        self.IF = _Radar_1_8(fi.read(76), estr)
        self.if_algo = _get_string(fi.read(8))
        self.PH = _Radar_1_8(fi.read(76), estr)
        # at line 122 of def

        self.ph_data_rcd, self.proc_product = struct.unpack(estr+'2i', fi.read(2*4))
        self.mission_text = _get_string(fi.read(8))
        self.ph_source, self.gps_week = struct.unpack(estr+'iI', fi.read(2*4))
        self.data_collect_reqh = _get_string(fi.read(14))
        self.res6 = fi.read(2)  # leave uninterpreted
        # at line 129

        self.grid_name = _get_string(fi.read(24))
        self.pix_val_linearity, self.complex_or_real, self.bits_per_magnitude, \
            self.bits_per_phase = struct.unpack(estr+'2i2H', fi.read(2*4+2*2))
        self.complex_order_type, self.pix_data_type, self.image_length, \
            self.image_cmp_scheme = struct.unpack(estr+'4i', fi.read(4*4))
        # at line 138

        self.apbo, self.asa_pitch, self.asa_squint, self.dsa_pitch, self.ira = \
            struct.unpack(estr+'5f', fi.read(5*4))
        self.rx_polarization = struct.unpack(estr+'2f', fi.read(2*4))
        self.tx_polarization = struct.unpack(estr+'2f', fi.read(2*4))
        self.v_avg = struct.unpack(estr+'3f', fi.read(3*4))
        self.apc_avg = struct.unpack(estr+'3f', fi.read(3*4))
        self.averaging_time, self.dgta = struct.unpack(estr+'2f', fi.read(2*4))
        # at line 153

        velocity_y, velocity_z = struct.unpack(estr+'2I', fi.read(2*4))
        self.velocity_y = _rescale_float(velocity_y, 1 << 16)
        self.velocity_z = _rescale_float(velocity_z, 1 << 16)

        self.ba, self.be = struct.unpack(estr+'2f', fi.read(2*4))
        self.az_geom_corr, self.range_geom_corr, self.az_win_fac_bw, \
            self.range_win_fac_bw = struct.unpack(estr+'2i2f', fi.read(4*4))
        self.az_win_id = _get_string(fi.read(48))
        self.range_win_id = _get_string(fi.read(48))
        # at line 163

        self.keep_out_viol_prcnt = struct.unpack(estr+'f', fi.read(4))[0]
        self.az_coeff = struct.unpack(estr+'6f', fi.read(6*4))
        self.pos_uncert = struct.unpack(estr+'3f', fi.read(3*4))
        self.nav_aiding_type = struct.unpack(estr+'i', fi.read(4))[0]
        self.two_dnl_phase_coeffs = struct.unpack(estr+'10f', fi.read(10*4))
        self.clutter_snr_thresh = struct.unpack(estr+'f', fi.read(4))[0]
        # at line 171

        self.elevation_coeff = struct.unpack(estr+'9f', fi.read(9*4))
        self.monopulse_coeff = struct.unpack(estr+'12f', fi.read(12*4))
        self.twist_pt_err_prcnt, self.tilt_pt_err_prcnt, self.az_pt_err_prcnt = \
            struct.unpack(estr+'3f', fi.read(3*4))
        sigma_n, self.take_num = struct.unpack(estr+'Ii', fi.read(2*4))
        self.sigma_n = _rescale_float(sigma_n, 1 << 23)

        self.if_sar_flags = struct.unpack(estr+'5i', fi.read(5*4))
        self.mu_threshold, self.gff_app_type = struct.unpack(estr+'fi', fi.read(2*4))
        self.res7 = fi.read(8)  # leave uninterpreted


#####################
# version 2 specific header parsing

# NB: I am only parsing the GSATIMG, APINFO, IFINFO, and GEOINFO blocks
#   because those are the only blocks referenced in the matlab that I
#   am mirroring


class _BlockHeader_2(object):
    """
    Read and interpret a block "sub"-header. This generically precedes every version
    2 data block, including the main file header
    """

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        self.name = _get_string(fi.read(16))
        self.major_version, self.minor_version = struct.unpack(estr+'HH', fi.read(2*2))
        what0 = fi.read(4)  # not sure what this is from looking at the matlab.
        self.size = struct.unpack(estr+'I', fi.read(4))[0]
        what1 = fi.read(4)  # not sure what this is from looking at the matlab.
        if (self.version == '2.0' and self.size == 64) or (self.version == '1.0' and self.size == 52):
            self.name = 'RADARINFO'  # fix known issue for some early version 2 GFF files

    @property
    def version(self):
        """
        str: The version
        """

        return '{}.{}'.format(self.major_version, self.minor_version)


# APINFO definitions
class _APInfo_1_0(object):
    """
    The APINFO block
    """
    serialized_length = 314

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        self.missionText = _get_string(fi.read(8))
        self.swVerNum = _get_string(fi.read(8))
        self.radarSerNum, self.phSource = struct.unpack(estr+'2I', fi.read(2*4))
        fi.read(2)
        self.phName = _get_string(fi.read(128))
        self.ctrFreq, self.wavelength = struct.unpack(estr+'2f', fi.read(2*4))
        self.rxPolarization, self.txPolarization = struct.unpack(estr+'2I', fi.read(2*4))
        self.azBeamWidth, self.elBeamWidth = struct.unpack(estr+'2f', fi.read(2*4))
        self.grazingAngle, self.squintAngle, self.gta, self.rngToBeamCtr = \
            struct.unpack(estr+'4f', fi.read(4*4))
        # line 16

        self.desSquint, self.desRng, self.desGTA, self.antPhaseCtrBear = \
            struct.unpack(estr+'4f', fi.read(4*4))
        self.ApTimeUTC = struct.unpack(estr+'6H', fi.read(6*2))
        self.flightTime, self.flightWeek = struct.unpack(estr+'2I', fi.read(2*4))
        self.chirpRate, self.xDistToStart = struct.unpack(estr+'2f', fi.read(2*4))
        self.momeasMode, self.radarMode = struct.unpack(estr+'2I', fi.read(2*4))
        # line 32

        self.rfoa = struct.unpack(estr+'f', fi.read(4))[0]
        self.apcVel = struct.unpack(estr+'3d', fi.read(3*8))
        self.apcLLH = struct.unpack(estr+'3d', fi.read(3*8))
        self.keepOutViol, self.gimStopTwist, self.gimStopTilt, self.gimStopAz = \
            struct.unpack(estr+'4f', fi.read(4*4))


class _APInfo_2_0(_APInfo_1_0):
    """
    The APINFO block
    """
    serialized_length = 318

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        _APInfo_1_0.__init__(self, fi, estr)
        self.apfdFactor = struct.unpack(estr+'i', fi.read(4))[0]


class _APInfo_3_0(_APInfo_2_0):
    """
    The APINFO block
    """
    serialized_length = 334

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        _APInfo_2_0.__init__(self, fi, estr)
        self.fastTimeSamples, self.adSampleFreq, self.apertureTime, \
            self.numPhaseHistories = struct.unpack(estr+'I2fI', fi.read(4*4))


class _APInfo_4_0(object):
    """
    The APINFO block
    """
    serialized_length = 418

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        # essentially the same as version 3, except the first two fields are longer
        self.missionText = _get_string(fi.read(50))
        self.swVerNum = _get_string(fi.read(50))
        self.radarSerNum, self.phSource = struct.unpack(estr+'2I', fi.read(2*4))
        fi.read(2)
        self.phName = _get_string(fi.read(128))
        self.ctrFreq, self.wavelength = struct.unpack(estr+'2f', fi.read(2*4))
        self.rxPolarization, self.txPolarization = struct.unpack(estr+'2I', fi.read(2*4))
        self.azBeamWidth, self.elBeamWidth = struct.unpack(estr+'2f', fi.read(2*4))
        self.grazingAngle, self.squintAngle, self.gta, self.rngToBeamCtr = \
            struct.unpack(estr+'4f', fi.read(4*4))
        # line 16

        self.desSquint, self.desRng, self.desGTA, self.antPhaseCtrBear = \
            struct.unpack(estr+'4f', fi.read(4*4))
        self.ApTimeUTC = struct.unpack(estr+'6H', fi.read(6*2))
        self.flightTime, self.flightWeek = struct.unpack(estr+'2I', fi.read(2*4))
        self.chirpRate, self.xDistToStart = struct.unpack(estr+'2f', fi.read(2*4))
        self.momeasMode, self.radarMode = struct.unpack(estr+'2I', fi.read(2*4))
        # line 32

        self.rfoa = struct.unpack(estr+'f', fi.read(4))[0]
        self.apcVel = struct.unpack(estr+'3d', fi.read(3*8))
        self.apcLLH = struct.unpack(estr+'3d', fi.read(3*8))
        self.keepOutViol, self.gimStopTwist, self.gimStopTilt, self.gimStopAz = \
            struct.unpack(estr+'4f', fi.read(4*4))

        self.apfdFactor = struct.unpack(estr+'i', fi.read(4))[0]
        self.fastTimeSamples, self.adSampleFreq, self.apertureTime, \
            self.numPhaseHistories = struct.unpack(estr+'I2fI', fi.read(4*4))


class _APInfo_5_0(_APInfo_4_0):
    """
    The APINFO block
    """
    serialized_length = 426

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        _APInfo_4_0.__init__(self, fi, estr)
        self.lightSpeed = struct.unpack(estr+'d', fi.read(8))[0]  # really?


class _APInfo_5_1(_APInfo_5_0):
    """
    The APINFO block
    """
    serialized_length = 430

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        _APInfo_5_0.__init__(self, fi, estr)
        self.delTanApAngle = struct.unpack(estr+'f', fi.read(4))[0]


class _APInfo_5_2(_APInfo_5_1):
    """
    The APINFO block
    """
    serialized_length = 434

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        _APInfo_5_1.__init__(self, fi, estr)
        self.metersInSampledDoppler = struct.unpack(estr+'f', fi.read(4))[0]


# IFINFO definitions
class _IFInfo_1_0(object):
    """
    Interpreter for IFInfo object
    """
    serialized_length = 514

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        self.procProduct = struct.unpack(estr+'I', fi.read(4))[0]
        fi.read(2)
        self.imgFileName = _get_string(fi.read(128))
        self.azResolution, self.rngResolution = struct.unpack(estr+'2f', fi.read(2*4))
        self.imgCalParam, self.sigmaN = struct.unpack(estr+'2f', fi.read(2*4))
        self.sampLocDCRow, self.sampLocDCCol = struct.unpack(estr+'2i', fi.read(2*4))
        self.ifAlgo = _get_string(fi.read(8))
        self.imgFlag = struct.unpack(estr+'i', fi.read(4))[0]
        self.azCoeff = struct.unpack(estr+'6f', fi.read(6*4))
        self.elCoeff = struct.unpack(estr+'9f', fi.read(9*4))
        self.azGeoCorrect, self.rngGeoCorrect = struct.unpack(estr+'2i', fi.read(2*4))
        self.wndBwFactAz, self.wndBwFactRng = struct.unpack(estr+'2f', fi.read(2*4))
        self.wndFncIdAz = _get_string(fi.read(48))
        self.wndFncIdRng = _get_string(fi.read(48))
        fi.read(2)
        self.cmtText = _get_string(fi.read(166))
        self.autoFocusInfo = struct.unpack(estr+'i', fi.read(4))[0]


class _IFInfo_2_0(_IFInfo_1_0):
    """
    Interpreter for IFInfo object - identical with version 2.1 and 2.2
    """
    serialized_length = 582

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        _IFInfo_1_0.__init__(self, fi, estr)
        self.rngFFTSize = struct.unpack(estr+'i', fi.read(4))[0]
        self.RangePaneFilterCoeff = struct.unpack(estr+'11f', fi.read(11*4))
        self.AzPreFilterCoeff = struct.unpack(estr+'5f', fi.read(5*4))


class _IFInfo_3_0(_IFInfo_2_0):
    """
    Interpreter for IFInfo object
    """
    serialized_length = 586

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        _IFInfo_2_0.__init__(self, fi, estr)
        self.afPeakQuadComp = struct.unpack(estr+'f', fi.read(4))[0]


# GEOINFO definitions
class _GeoInfo_1(object):
    """
    Interpreter for GeoInfo object - note that versions 1.0 and 1.1 are identical
    """
    serialized_length = 52

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        self.imagePlane = struct.unpack(estr+'i', fi.read(4))[0]
        self.rangePixSpacing, self.desiredGrazAng, self.azPixSpacing = \
            struct.unpack(estr+'3f', fi.read(3*4))
        self.patchCtrLLH = struct.unpack(estr+'3d', fi.read(3*8))
        self.pixLocImCtrRow, self.pixLocImCtrCol = struct.unpack(estr+'2I', fi.read(2*4))
        self.imgRotAngle = struct.unpack(estr+'f', fi.read(4))[0]


# GSATIMG definition

def _get_complex_domain_code(code_int):
    # type: (int) -> str
    if code_int in [0, 3]:
        return 'IQ'
    elif code_int in [1, 4]:
        return 'QI'
    elif code_int in [2, 5]:
        return 'MP'
    elif code_int == 6:
        return 'PM'
    elif code_int == 7:
        return 'M'
    elif code_int == 8:
        return 'P'
    else:
        raise ValueError('Got unexpected code `{}`'.format(code_int))


def _get_band_order(code_int):
    # type: (int) -> str
    if code_int in [0, 1, 2, 7, 8]:
        return 'interleaved'
    elif code_int in [3, 4, 5, 6]:
        return 'sequential'
    else:
        raise ValueError('Got unexpected code `{}`'.format(code_int))


class _PixelFormat(object):
    """
    Interpreter for pixel format object
    """

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        self.comp0_bitSize, self.comp0_dataType = struct.unpack(estr+'HI', fi.read(2+4))
        self.comp1_bitSize, self.comp1_dataType = struct.unpack(estr+'HI', fi.read(2+4))
        self.cmplxDomain, self.numComponents = struct.unpack(estr+'Ii', fi.read(2*4))


class _GSATIMG_2(object):
    """
    Interpreter for the GSATIMG object
    """
    serialized_length = 82

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        self.endian = struct.unpack(estr+'I', fi.read(4))[0]
        fi.read(2)
        self.imageCreator = _get_string(fi.read(24))
        self.rangePixels, self.azPixels = struct.unpack(estr+'2I', fi.read(2*4))
        self.pixOrder, self.imageLengthBytes, self.imageCompressionScheme, \
            self.pixDataType = struct.unpack(estr+'4I', fi.read(4*4))
        self.pixelFormat = _PixelFormat(fi, estr)
        self.pixValLin, self.autoScaleFac = struct.unpack(estr+'if', fi.read(2*4))

        complex_domain = _get_complex_domain_code(self.pixelFormat.cmplxDomain)
        if complex_domain not in ['IQ', 'QI', 'MP', 'PM']:
            raise ValueError('We got unsupported complex domain `{}`'.format(complex_domain))


# combined GFF version 2 header collection
def _check_serialization(
        block_header: _BlockHeader_2,
        expected_length: int) -> None:
    if block_header.size == expected_length:
        return

    raise ValueError(
        'Got `{}` block of version `{}` and serialized length {},\n\t'
        'but expected serialized length {}'.format(
            block_header.name, block_header.version, block_header.size, expected_length))


class _GFFHeader_2(object):
    """
    Interpreter for the GFF version 2.* header
    """

    __slots__ = (
        'file_object', 'estr', '_gsat_img', '_ap_info', '_if_info', '_geo_info',
        '_image_header', '_image_offset')

    def __init__(self, fi, estr):
        """

        Parameters
        ----------
        fi : BinaryIO
        estr : str
            The endianness string for format interpretation, one of `['<', '>']`
        """

        self._gsat_img = None
        self._ap_info = None
        self._if_info = None
        self._geo_info = None
        self._image_header = None
        self._image_offset = None
        self.file_object = fi
        self.estr = estr

        # extract the initial file location
        init_location = fi.tell()

        # go to the begining of the file
        fi.seek(0, os.SEEK_SET)
        gsat_header = _BlockHeader_2(fi, estr)
        self._gsat_img = _GSATIMG_2(fi, estr)

        while True:
            block_header = _BlockHeader_2(fi, estr)
            if block_header.name == 'IMAGEDATA':
                self._image_header = block_header
                self._image_offset = fi.tell()
                break
            elif block_header.name == 'APINFO':
                self._parse_apinfo(fi, estr, block_header)
            elif block_header.name == 'IFINFO':
                self._parse_ifinfo(fi, estr, block_header)
            elif block_header.name == 'GEOINFO':
                self._parse_geoinfo(fi, estr, block_header)
            else:
                # we are not parsing this block, so just skip it
                fi.seek(block_header.size, os.SEEK_CUR)

        # return to the initial file location
        fi.seek(init_location, os.SEEK_SET)
        self._check_valid(gsat_header)

    @property
    def gsat_img(self) -> _GSATIMG_2:
        return self._gsat_img

    @property
    def ap_info(self) -> Union[_APInfo_1_0, _APInfo_2_0, _APInfo_3_0, _APInfo_4_0, _APInfo_5_0, _APInfo_5_1, _APInfo_5_2]:
        return self._ap_info

    @property
    def if_info(self) -> Union[_IFInfo_1_0, _IFInfo_2_0, _IFInfo_3_0]:
        return self._if_info

    @property
    def geo_info(self) -> _GeoInfo_1:
        return self._geo_info

    @property
    def image_header(self) -> _BlockHeader_2:
        return self._image_header

    @property
    def image_offset(self) -> int:
        return self._image_offset

    def _parse_apinfo(self, fi, estr, block_header) -> None:
        if block_header.name != 'APINFO':
            return

        if block_header.major_version == 1:
            _check_serialization(block_header, _APInfo_1_0.serialized_length)
            self._ap_info = _APInfo_1_0(fi, estr)
        elif block_header.major_version == 2:
            _check_serialization(block_header, _APInfo_2_0.serialized_length)
            self._ap_info = _APInfo_2_0(fi, estr)
        elif block_header.major_version == 3:
            _check_serialization(block_header, _APInfo_3_0.serialized_length)
            self._ap_info = _APInfo_3_0(fi, estr)
        elif block_header.major_version == 4:
            _check_serialization(block_header, _APInfo_4_0.serialized_length)
            self._ap_info = _APInfo_4_0(fi, estr)
        elif block_header.major_version == 5:
            if block_header.minor_version == 0:
                _check_serialization(block_header, _APInfo_5_0.serialized_length)
                self._ap_info = _APInfo_5_0(fi, estr)
            elif block_header.minor_version == 1:
                _check_serialization(block_header, _APInfo_5_1.serialized_length)
                self._ap_info = _APInfo_5_1(fi, estr)
            elif block_header.minor_version == 2:
                _check_serialization(block_header, _APInfo_5_2.serialized_length)
                self._ap_info = _APInfo_5_2(fi, estr)
        else:
            raise ValueError(
                'Could not parse required `{}` block version `{}`'.format(
                    block_header.name, block_header.version))

    def _parse_ifinfo(self, fi, estr, block_header) -> None:
        if block_header.name != 'IFINFO':
            return

        if block_header.major_version == 1:
            _check_serialization(block_header, _IFInfo_1_0.serialized_length)
            self._if_info = _IFInfo_1_0(fi, estr)
        elif block_header.major_version == 2:
            _check_serialization(block_header, _IFInfo_2_0.serialized_length)
            self._if_info = _IFInfo_2_0(fi, estr)
        elif block_header.major_version == 3:
            _check_serialization(block_header, _IFInfo_3_0.serialized_length)
            self._if_info = _IFInfo_3_0(fi, estr)
        else:
            raise ValueError(
                'Could not parse required `{}` block version `{}`'.format(
                    block_header.name, block_header.version))

    def _parse_geoinfo(self, fi, estr, block_header) -> None:
        if block_header.name != 'GEOINFO':
            return

        _check_serialization(block_header, _GeoInfo_1.serialized_length)
        self._geo_info = _GeoInfo_1(fi, estr)

    def _check_valid(self, gsat_header) -> None:
        # ensure that the required elements are all set
        valid = True
        if self._ap_info is None:
            valid = False
            logger.error(
                'GFF version {} file did not present APINFO block'.format(
                    gsat_header.version))
        if self._if_info is None:
            valid = False
            logger.error(
                'GFF version {} file did not present IFINFO block'.format(
                    gsat_header.version))
        if self._geo_info is None:
            valid = False
            logger.error(
                'GFF version {} file did not present GEOINFO block'.format(
                    gsat_header.version))
        if not valid:
            raise ValueError('GFF file determined to be invalid')

    def get_arp_vel(self) -> numpy.ndarray:
        """
        Gets the aperture velocity in ECF coordinates

        Returns
        -------
        numpy.ndarray
        """

        # get the aperture velocity in its native frame of reference (rotated ENU)
        arp_vel_orig = numpy.array(self.ap_info.apcVel, dtype='float64')
        # TODO: arp_vel_orig is in what coordinate system? Rick said "rotated ENU", wrt gta?
        # gets the angle wrt to True North for the radar frame of reference
        angle = numpy.deg2rad(self.ap_info.rfoa)
        cosine, sine = numpy.cos(angle), numpy.sin(angle)
        # construct the NED velocity vector
        transform = numpy.array([[cosine, -sine, 0], [sine, cosine, 0], [0, 0, -1]], dtype='float64')
        ned_velocity = transform.dot(arp_vel_orig)
        # convert to ECF
        orp = geodetic_to_ecf(self.ap_info.apcLLH, ordering='latlon')
        out = ned_to_ecf(ned_velocity, orp, absolute_coords=False)
        return out


####################
# object for creation of sicd structure from GFF header object

def _get_wgt(str_in: str) -> Optional[WgtTypeType]:
    if str_in == '':
        return None

    elements = str_in.split()
    win_name = elements[0].upper()
    parameters = None
    if win_name == 'TAYLOR':
        if len(elements) < 2:
            raise ValueError('Got unparseable window definition `{}`'.format(str_in))
        params = elements[1].split(',')
        if len(params) != 2:
            raise ValueError('Got unparseable window definition `{}`'.format(str_in))
        parameters = {'SLL': params[0].strip(), 'NBAR': params[1].strip()}
    return WgtTypeType(
        WindowName=win_name,
        Parameters=parameters)


def _get_polarization_string(int_value: int) -> Optional[str]:
    if int_value == 0:
        return 'H'
    elif int_value == 1:
        return 'V'
    elif int_value == 2:
        return 'LHC'
    elif int_value == 3:
        return 'RHC'
    elif int_value in [4, 5]:
        # TODO: according to their enum, we have 4 -> "T" and 5 -> "P"
        #   what does that mean?
        return 'OTHER'
    else:
        return 'UNKNOWN'


def _get_tx_rcv_polarization(tx_pol_int: int, rcv_pol_int: int) -> Tuple[str, str]:
    tx_pol = _get_polarization_string(tx_pol_int)
    rcv_pol = _get_polarization_string(rcv_pol_int)
    if tx_pol in ['OTHER', 'UNKNOWN'] or rcv_pol in ['OTHER', 'UNKNOWN']:
        tx_rcv_pol = 'OTHER'
    else:
        tx_rcv_pol = '{}:{}'.format(tx_pol, rcv_pol)
    return tx_pol, tx_rcv_pol


class _GFFInterpreter(object):
    """
    Extractor for the sicd details
    """

    def get_sicd(self) -> SICDType:
        """
        Gets the SICD structure.

        Returns
        -------
        SICDType
        """

        raise NotImplementedError

    def get_data_segment(self) -> DataSegment:
        """
        Gets the chipper for reading the data.

        Returns
        -------
        DataSegment
        """

        raise NotImplementedError

    def clean_up(self) -> None:
        return


class _GFFInterpreter1(_GFFInterpreter):
    """
    Extractor of SICD structure and parameters from gff_header_1*
    object
    """

    def __init__(self, header: Union[_GFFHeader_1_6, _GFFHeader_1_8]):
        """

        Parameters
        ----------
        header : _GFFHeader_1_6|_GFFHeader_1_8
        """

        self.header = header
        if self.header.image_type == 0:
            raise ValueError(
                'ImageType indicates a magnitude only image, which is incompatible with SICD')

    def get_sicd(self) -> SICDType:
        def get_collection_info() -> CollectionInfoType:
            core_name = self.header.image_name.replace(':', '_')
            return CollectionInfoType(
                CoreName=core_name,
                CollectType='MONOSTATIC',
                RadarMode=RadarModeType(
                    ModeType='SPOTLIGHT'),
                Classification='UNCLASSIFIED')

        def get_image_creation() -> ImageCreationType:
            from sarpy.__about__ import __version__
            from datetime import datetime
            return ImageCreationType(
                Application=self.header.creator,
                DateTime=numpy.datetime64(datetime(*self.header.date_time)),
                Site='Unknown',
                Profile='sarpy {}'.format(__version__))

        def get_image_data() -> ImageDataType:
            return ImageDataType(
                PixelType='RE32F_IM32F',
                NumRows=num_rows,
                NumCols=num_cols,
                FullImage=(num_rows, num_cols),
                FirstRow=0,
                FirstCol=0,
                SCPPixel=(scp_row, scp_col))

        def get_geo_data() -> GeoDataType:
            return GeoDataType(
                SCP=SCPType(
                    LLH=[self.header.srp_lat, self.header.srp_lon, self.header.srp_alt]))

        def get_grid() -> GridType:
            image_plane = 'GROUND' if self.header.image_plane == 0 else 'SLANT'
            # we presume that image_plane in [0, 1]

            row_ss = self.header.range_pixel_size
            col_ss = self.header.azimuth_pixel_size
            row_bw = 1./row_ss
            col_bw = 1./col_ss
            if self.header.version == '1.8':
                if self.header.range_win_fac_bw > 0:
                    row_bw = self.header.range_win_fac_bw/row_ss
                if self.header.az_win_fac_bw > 0:
                    col_bw = self.header.az_win_fac_bw/col_ss

            row = DirParamType(
                Sgn=-1,
                SS=row_ss,
                ImpRespWid=self.header.range_resolution,
                ImpRespBW=row_bw,
                DeltaK1=0.5*row_bw,
                DeltaK2=-0.5*row_bw,
                WgtType=_get_wgt(
                    self.header.range_win_id if self.header.version == '1.8' else ''),
                DeltaKCOAPoly=[[0, ], ]
            )

            col = DirParamType(
                Sgn=-1,
                SS=col_ss,
                ImpRespWid=self.header.az_resolution,
                ImpRespBW=col_bw,
                DeltaK1=0.5*col_bw,
                DeltaK2=-0.5*col_bw,
                WgtType=_get_wgt(
                    self.header.az_win_id if self.header.version == '1.8' else ''),
                DeltaKCOAPoly=[[0, ], ]
            )

            return GridType(
                ImagePlane=image_plane,
                Type='PLANE',
                Row=row,
                Col=col)

        def get_scpcoa() -> SCPCOAType:
            side_of_track = 'L' if self.header.squint < 0 else 'R'

            apc_llh = numpy.array(
                [self.header.apc_lat, self.header.apc_lon, self.header.apc_alt],
                dtype='float64')
            if numpy.all(apc_llh == 0):
                arp_pos = None
            else:
                arp_pos = geodetic_to_ecf(apc_llh, ordering='latlon')

            return SCPCOAType(
                ARPPos=arp_pos,
                GrazeAng=self.header.graze_angle,
                SideOfTrack=side_of_track)

        num_rows = self.header.range_count
        num_cols = self.header.azimuth_count
        scp_row = int(0.5*num_rows)
        scp_col = int(0.5*num_cols)

        collection_info = get_collection_info()
        image_creation = get_image_creation()
        image_data = get_image_data()
        geo_data = get_geo_data()
        grid = get_grid()
        scpcoa = get_scpcoa()

        return SICDType(
            CollectionInfo=collection_info,
            ImageCreation=image_creation,
            ImageData=image_data,
            GeoData=geo_data,
            Grid=grid,
            SCPCOA=scpcoa)

    def get_data_segment(self) -> DataSegment:
        if self.header.bits_per_phase not in [8, 16, 32]:
            raise ValueError('Got unexpected bits per phase {}'.format(self.header.bits_per_phase))
        if self.header.bits_per_magnitude not in [8, 16, 32]:
            raise ValueError('Got unexpected bits per phase {}'.format(self.header.bits_per_magnitude))

        raw_dtype = numpy.dtype('{}u{}'.format(self.header.estr, int(self.header.bits_per_phase/8)))
        raw_shape = (self.header.range_count, self.header.azimuth_count, 2)
        reverse_axes = (0, 1)
        if self.header.row_major:
            transpose_axes = (1, 0, 2)
            formatted_shape = (raw_shape[1], raw_shape[0])
        else:
            transpose_axes = None
            formatted_shape = raw_shape[:2]

        if self.header.image_type == 1:
            # phase/magnitude which is integer
            return NumpyMemmapSegment(
                self.header.file_object, self.header.header_length,
                raw_dtype, raw_shape, formatted_dtype='complex64', formatted_shape=formatted_shape,
                reverse_axes=reverse_axes, transpose_axes=transpose_axes,
                format_function=ComplexFormatFunction(raw_dtype, order='PM', band_dimension=2),
                close_file=True)
        else:
            raise ValueError('Got unsupported image type `{}`'.format(self.header.image_type))


def _get_numpy_dtype(data_type_int: int) -> str:
    if data_type_int == 0:
        return 'u1'
    elif data_type_int == 1:
        return 'u2'
    elif data_type_int == 2:
        return 'u4'
    elif data_type_int == 3:
        return 'u8'
    elif data_type_int == 4:
        return 'i1'
    elif data_type_int == 5:
        return 'i2'
    elif data_type_int == 6:
        return 'i4'
    elif data_type_int == 7:
        return 'i8'
    elif data_type_int == 8:
        return 'f4'
    elif data_type_int == 9:
        return 'f8'
    else:
        raise ValueError('Got unsupported data type code `{}`'.format(data_type_int))


class _GFFInterpreter2(_GFFInterpreter):
    """
    Extractor of SICD structure and parameters from GFFHeader_2 object
    """

    def __init__(self, header: _GFFHeader_2):
        """

        Parameters
        ----------
        header : _GFFHeader_2
        """

        self.header = header
        self._cached_files = []
        if self.header.gsat_img.pixelFormat.numComponents != 2:
            raise ValueError(
                'The pixel format indicates that the number of components is `{}`, '
                'which is not supported for a complex image'.format(
                    self.header.gsat_img.pixelFormat.numComponents))

    def get_sicd(self) -> SICDType:
        def get_collection_info() -> CollectionInfoType:
            core_name = self.header.ap_info.phName  # TODO: double check this...
            return CollectionInfoType(
                CollectorName=self.header.ap_info.missionText,
                CoreName=core_name,
                CollectType='MONOSTATIC',
                RadarMode=RadarModeType(
                    ModeType='SPOTLIGHT'),
                Classification='UNCLASSIFIED')

        def get_image_creation() -> ImageCreationType:
            from sarpy.__about__ import __version__
            application = '{} {}'.format(self.header.gsat_img.imageCreator, self.header.ap_info.swVerNum)
            date_time = None  # todo: really?
            return ImageCreationType(
                Application=application,
                DateTime=date_time,
                Site='Unknown',
                Profile='sarpy {}'.format(__version__))

        def get_image_data() -> ImageDataType:
            pix_data_type = self.header.gsat_img.pixDataType
            amp_table = None
            if pix_data_type == 12:
                pixel_type = 'AMP8I_PHS8I'
                amp_table = numpy.arange(256, dtype='float64')
            elif pix_data_type in [1, 3, 4, 6, 8, 9, 10, 11]:
                pixel_type = 'RE32F_IM32F'
            elif pix_data_type in [2, 7]:
                pixel_type = 'RE16I_IM16I'
            else:
                raise ValueError('Unhandled pixTypeData value `{}`'.format(pix_data_type))

            return ImageDataType(
                PixelType=pixel_type,
                AmpTable=amp_table,
                NumRows=num_rows,
                NumCols=num_cols,
                FullImage=(num_rows, num_cols),
                FirstRow=0,
                FirstCol=0,
                SCPPixel=(scp_row, scp_col))

        def get_geo_data() -> GeoDataType:
            return GeoDataType(SCP=SCPType(ECF=scp))

        def get_grid() -> GridType:
            image_plane = 'GROUND' if self.header.geo_info.imagePlane == 0 else 'SLANT'
            # we presume that image_plane in [0, 1]

            # derive row/col uvect
            ground_uvec = wgs_84_norm(scp)

            urng = scp - arp_pos  # unit vector for row in the slant plane
            urng /= numpy.linalg.norm(urng)
            if image_plane == 'GROUND':
                row_uvec = urng - numpy.dot(urng, ground_uvec)*ground_uvec
                row_uvec /= numpy.linalg.norm(row_uvec)
            else:
                row_uvec = urng

            col_uvec = arp_vel/numpy.linalg.norm(arp_vel)
            if self.header.ap_info.squintAngle < 0:
                col_uvec *= -1

            # verify that my orientation makes some sense
            dumb_check = ground_uvec.dot(numpy.cross(row_uvec, col_uvec))
            if dumb_check <= 0:
                raise ValueError(
                    'The range vector, velocity vector, and squint angle have '
                    'incompatible orientations')

            parallel_component = numpy.dot(row_uvec, col_uvec)
            if numpy.abs(parallel_component) > 1e-7:
                col_uvec = col_uvec - parallel_component*row_uvec
                col_uvec /= numpy.linalg.norm(col_uvec)

            row_ss = self.header.geo_info.rangePixSpacing
            row_bw = self.header.if_info.wndBwFactRng/self.header.if_info.rngResolution
            row_delta_kcoa_constant = 0.5*(1 - (self.header.if_info.sampLocDCRow/int(0.5*num_rows)))/row_ss
            row = DirParamType(
                Sgn=-1,
                SS=row_ss,
                UVectECF=row_uvec,
                ImpRespWid=self.header.if_info.rngResolution,
                ImpRespBW=row_bw,
                KCtr=2*center_frequency/speed_of_light,
                DeltaK1=0.5*row_bw,
                DeltaK2=-0.5*row_bw,
                WgtType=_get_wgt(self.header.if_info.wndFncIdRng),
                DeltaKCOAPoly=[[row_delta_kcoa_constant, ], ])

            col_ss = self.header.geo_info.azPixSpacing
            col_bw = self.header.if_info.wndBwFactAz/self.header.if_info.azResolution
            col_delta_kcoa_constant = 0.5*(1 - (self.header.if_info.sampLocDCCol/int(0.5*num_cols)))/col_ss
            col = DirParamType(
                Sgn=-1,
                SS=col_ss,
                UVectECF=col_uvec,
                ImpRespWid=self.header.if_info.azResolution,
                ImpRespBW=col_bw,
                KCtr=0,
                DeltaK1=0.5*col_bw,
                DeltaK2=-0.5*col_bw,
                WgtType=_get_wgt(self.header.if_info.wndFncIdAz),
                DeltaKCOAPoly=[[col_delta_kcoa_constant, ], ])

            return GridType(
                ImagePlane=image_plane,
                Type=grid_type,
                Row=row,
                Col=col)

        def get_scpcoa() -> SCPCOAType:
            return SCPCOAType(
                ARPPos=arp_pos,
                ARPVel=arp_vel,
                SCPTime=0.5*collect_duration)

        def get_timeline() -> TimelineType:
            try:
                # only exists for APINFO version 3 and above
                ipp_end = self.header.ap_info.numPhaseHistories
                ipp = [IPPSetType(
                    TStart=0,
                    TEnd=collect_duration,
                    IPPStart=0,
                    IPPEnd=ipp_end,
                    IPPPoly=[0, (ipp_end + 1)/collect_duration]), ]
            except AttributeError:
                ipp = None
            return TimelineType(
                CollectStart=start_time,
                CollectDuration=collect_duration,
                IPP=ipp)

        def get_radar_collection() -> RadarCollectionType:

            try:
                sample_rate = self.header.ap_info.adSampleFreq
                pulse_length = float(self.header.ap_info.fastTimeSamples)/sample_rate
                waveform = [
                    WaveformParametersType(ADCSampleRate=sample_rate, TxPulseLength=pulse_length), ]
            except AttributeError:
                waveform = None

            rcv_channels = [ChanParametersType(TxRcvPolarization=tx_rcv_pol, index=1), ]

            return RadarCollectionType(
                TxFrequency=(center_frequency-0.5*band_width, center_frequency+0.5*band_width),
                Waveform=waveform,
                TxPolarization=tx_pol,
                RcvChannels=rcv_channels)

        def get_image_formation() -> ImageFormationType:
            return ImageFormationType(
                RcvChanProc=RcvChanProcType(ChanIndices=[1, ]),
                TxRcvPolarizationProc=tx_rcv_pol,
                TxFrequencyProc=(
                    center_frequency-0.5*band_width,
                    center_frequency+0.5*band_width),
                TStartProc=0,
                TEndProc=collect_duration,
                ImageFormAlgo=image_form_algo,
                STBeamComp='NO',
                ImageBeamComp='NO',
                AzAutofocus='NO',
                RgAutofocus='NO')

        def repair_scpcoa() -> None:
            # call after deriving the sicd fields
            if out_sicd.SCPCOA.GrazeAng is None:
                out_sicd.SCPCOA.GrazeAng = self.header.ap_info.grazingAngle
            if out_sicd.SCPCOA.IncidenceAng is None:
                out_sicd.SCPCOA.IncidenceAng = 90 - out_sicd.SCPCOA.GrazeAng
            if out_sicd.SCPCOA.SideOfTrack is None:
                out_sicd.SCPCOA.SideOfTrack = 'L' if self.header.ap_info.squintAngle < 0 else 'R'

        def populate_radiometric() -> None:
            # call after deriving the sicd fields
            rcs_constant = self.header.if_info.imgCalParam**2
            radiometric = RadiometricType(RCSSFPoly=[[rcs_constant, ]])
            # noinspection PyProtectedMember
            radiometric._derive_parameters(out_sicd.Grid, out_sicd.SCPCOA)
            if radiometric.SigmaZeroSFPoly is not None:
                noise_constant = self.header.if_info.sigmaN - 10*numpy.log10(radiometric.SigmaZeroSFPoly[0, 0])
                radiometric.NoiseLevel = NoiseLevelType_(
                    NoiseLevelType='ABSOLUTE',
                    NoisePoly=[[noise_constant, ]])
            out_sicd.Radiometric = radiometric

        num_rows = self.header.gsat_img.rangePixels
        num_cols = self.header.gsat_img.azPixels
        scp_row = self.header.geo_info.pixLocImCtrRow
        scp_col = self.header.geo_info.pixLocImCtrCol

        collect_duration = self.header.ap_info.apertureTime
        scp_time_utc_us = numpy.datetime64(datetime(*self.header.ap_info.ApTimeUTC), 'us').astype('int64')
        start_time = (scp_time_utc_us - int(0.5*collect_duration*1e6)).astype('datetime64[us]')
        tx_pol, tx_rcv_pol = _get_tx_rcv_polarization(
            self.header.ap_info.txPolarization, self.header.ap_info.rxPolarization)
        center_frequency = self.header.ap_info.ctrFreq
        band_width = 0.0  # TODO: is this defined anywhere?

        scp = geodetic_to_ecf(self.header.geo_info.patchCtrLLH)
        arp_llh = self.header.ap_info.apcLLH
        arp_pos = geodetic_to_ecf(arp_llh, ordering='latlon')
        arp_vel = self.header.get_arp_vel()

        if self.header.if_info.ifAlgo in ['PFA', 'OSAPF']:
            # if self.header.if_info.ifAlgo == 'PFA':
            image_form_algo = 'PFA'
            grid_type = 'RGAZIM'
        else:
            image_form_algo = 'OTHER'
            grid_type = 'PLANE'

        collection_info = get_collection_info()
        image_creation = get_image_creation()
        image_data = get_image_data()
        geo_data = get_geo_data()
        scp = geo_data.SCP.ECF.get_array()

        grid = get_grid()
        scpcoa = get_scpcoa()
        timeline = get_timeline()
        radar_collection = get_radar_collection()
        image_formation = get_image_formation()

        out_sicd = SICDType(
            CollectionInfo=collection_info,
            ImageCreation=image_creation,
            ImageData=image_data,
            GeoData=geo_data,
            Grid=grid,
            SCPCOA=scpcoa,
            Timeline=timeline,
            RadarCollection=radar_collection,
            ImageFormation=image_formation)

        out_sicd.derive()
        repair_scpcoa()
        populate_radiometric()
        out_sicd.populate_rniirs(override=False)
        return out_sicd

    def _get_size_and_symmetry(self) -> Tuple[Tuple[int, int], Tuple[int, ...], bool]:
        if self.header.gsat_img.pixOrder == 0:
            # in range consecutive order, opposite from a SICD
            data_size = (self.header.gsat_img.azPixels, self.header.gsat_img.rangePixels)
            reverse_axes = (0, 1)
            transpose_axes = True
        elif self.header.gsat_img.pixOrder == 1:
            # in azimuth consecutive order, like a SICD
            data_size = (self.header.gsat_img.rangePixels, self.header.gsat_img.azPixels)
            reverse_axes = (0, 1)
            transpose_axes = False
        else:
            raise ValueError('Got unexpected pixel order `{}`'.format(self.header.gsat_img.pixOrder))
        return data_size, reverse_axes, transpose_axes

    def _check_image_validity(self, band_order: str) -> None:
        if self.header.gsat_img.pixelFormat.numComponents != 2:
            raise ValueError(
                'Got unexpected number of components `{}`'.format(
                    self.header.gsat_img.pixelFormat.numComponents))

        image_compression_scheme = self.header.gsat_img.imageCompressionScheme
        if image_compression_scheme in [1, 3]:
            if band_order == 'sequential':
                raise ValueError(
                    'GFF with sequential bands and jpeg or jpeg 2000 compression currently unsupported.')
            if PIL is None:
                raise ValueError(
                    'The GFF image is compressed using jpeg or jpeg 2000 compression, '
                    'and decompression requires the PIL library')

    def _extract_zlib_image(self) -> str:
        if self.header.gsat_img.imageCompressionScheme != 2:
            raise ValueError('The image is not zlib compressed')
        self.header.file_object.seek(self.header.image_offset, os.SEEK_SET)
        data_bytes = zlib.decompress(self.header.file_object.read(self.header.image_header.size))
        fi, path_name = mkstemp(suffix='.sarpy_cache', text=False)
        os.close(fi)
        self._cached_files.append(path_name)
        logger.info('Created cached file {} for decompressed data'.format(path_name))
        with open(path_name, 'wb') as the_file:
            the_file.write(data_bytes)
        logger.info('Filled cached file {}'.format(path_name))
        return path_name

    def _extract_pil_image(
            self,
            band_order: str,
            data_size: Tuple[int, int]) -> str:
        if band_order == 'sequential':
            raise ValueError(
                'GFF with sequential bands and jpeg or jpeg 2000 compression currently unsupported.')
        our_memmap = MemMap(self.header.file_object.name, self.header.image_header.size, self.header.image_offset)
        # noinspection PyUnresolvedReferences
        img = PIL.Image.open(our_memmap)  # this is a lazy operation
        # dump the extracted image data out to a temp file
        fi, path_name = mkstemp(suffix='.sarpy_cache', text=False)
        os.close(fi)
        self._cached_files.append(path_name)
        logger.info('Created cached file {} for decompressed data'.format(path_name))
        data = numpy.asarray(img)  # create our numpy array from the PIL Image
        if data.shape[:2] != data_size:
            raise ValueError(
                'Naively decompressed data of shape {}, but expected ({}, {}, {}).'.format(
                    data.shape, data_size[0], data_size[1], 2))
        mem_map = numpy.memmap(path_name, dtype=data.dtype, mode='w+', offset=0, shape=data.shape)
        mem_map[:] = data
        # clean up this memmap and file overhead
        del mem_map
        logger.info('Filled cached file {}'.format(path_name))
        return path_name

    def _get_interleaved_segment(self) -> DataSegment:
        complex_domain = _get_complex_domain_code(self.header.gsat_img.pixelFormat.cmplxDomain)
        if complex_domain not in ['IQ', 'QI', 'MP', 'PM']:
            raise ValueError('Got unsupported complex domain `{}`'.format(complex_domain))

        if self.header.gsat_img.pixelFormat.comp0_dataType != self.header.gsat_img.pixelFormat.comp1_dataType:
            raise ValueError(
                'GFF with interleaved bands with the two components of different data types.\n\t'
                'This is not currently unsupported.')
        raw_dtype = _get_numpy_dtype(self.header.gsat_img.pixelFormat.comp0_dataType)

        data_size, reverse_axes, transpose = self._get_size_and_symmetry()
        raw_shape = data_size + (2, )
        if transpose:
            transpose_axes = (1, 0, 2)
            formatted_shape = (data_size[1], data_size[0])
        else:
            transpose_axes = None
            formatted_shape = data_size

        image_compression_scheme = self.header.gsat_img.imageCompressionScheme
        if image_compression_scheme == 0:
            # no compression
            the_file = self.header.file_object
            data_offset = self.header.image_offset
        elif image_compression_scheme in [1, 3]:
            # jpeg or jpeg 2000 compression
            the_file = self._extract_pil_image('interleaved', data_size)
            data_offset = 0
        elif image_compression_scheme == 2:
            # zlib compression
            the_file = self._extract_zlib_image()
            data_offset = 0
        else:
            raise ValueError('Got unhandled image compression scheme code `{}`'.format(image_compression_scheme))

        return NumpyMemmapSegment(
            the_file, data_offset, raw_dtype, raw_shape,
            formatted_dtype='complex64', formatted_shape=formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=ComplexFormatFunction(raw_dtype, complex_domain, band_dimension=2))

    def _get_sequential_segment(self) -> DataSegment:
        complex_domain = _get_complex_domain_code(self.header.gsat_img.pixelFormat.cmplxDomain)
        if complex_domain not in ['IQ', 'QI', 'MP', 'PM']:
            raise ValueError('Got unsupported complex domain `{}`'.format(complex_domain))

        if self.header.gsat_img.pixelFormat.comp0_dataType != self.header.gsat_img.pixelFormat.comp1_dataType:
            raise ValueError(
                'GFF with sequential bands has the two components with different data types.\n\t'
                'This is not currently unsupported.')

        raw_dtype = numpy.dtype(_get_numpy_dtype(self.header.gsat_img.pixelFormat.comp0_dataType))
        data_size, reverse_axes, transpose = self._get_size_and_symmetry()
        raw_shape = (2, ) + data_size
        if reverse_axes is not None:
            reverse_axes = tuple(1+entry for entry in reverse_axes)
        if transpose:
            transpose_axes = (2, 1, 0)
            formatted_shape = (data_size[1], data_size[0])
        else:
            transpose_axes = (1, 2, 0)
            formatted_shape = data_size

        image_compression_scheme = self.header.gsat_img.imageCompressionScheme
        if image_compression_scheme == 0:
            # no compression
            the_file = self.header.file_object
            main_offset = self.header.image_offset
        elif image_compression_scheme == 2:
            the_file = self._extract_zlib_image()
            main_offset = 0
        else:
            raise ValueError('Unhandled image compression scheme `{}`'.format(image_compression_scheme))

        return NumpyMemmapSegment(
            the_file, main_offset, raw_dtype, raw_shape,
            formatted_dtype='complex64', formatted_shape=formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=ComplexFormatFunction(raw_dtype, complex_domain, band_dimension=0))

    def get_data_segment(self) -> DataSegment:
        band_order = _get_band_order(self.header.gsat_img.pixelFormat.cmplxDomain)
        self._check_image_validity(band_order)

        if band_order == 'interleaved':
            return self._get_interleaved_segment()
        elif band_order == 'sequential':
            return self._get_sequential_segment()
        else:
            raise ValueError('Unhandled band order `{}`'.format(band_order))

    def clean_up(self) -> None:
        try:
            if self._cached_files is not None:
                for fil in self._cached_files:
                    if os.path.exists(fil):
                        # noinspection PyBroadException
                        try:
                            os.remove(fil)
                            logger.info('Deleted cached file {}'.format(fil))
                        except Exception:
                            logger.error(
                                'Error in attempt to delete cached file {}.\n\t'
                                'Manually delete this file'.format(fil), exc_info=True)
            self._cached_files = None
        except AttributeError:
            return

    def __del__(self):
        """
        Clean up any cached files.

        Returns
        -------
        None
        """

        self.clean_up()


####################
# the actual reader implementation

class GFFDetails(object):
    __slots__ = (
        '_file_name', '_file_object', '_close_after',
        '_endianness', '_major_version', '_minor_version',
        '_header', '_interpreter')

    def __init__(self, file_name: str):
        """

        Parameters
        ----------
        file_name : str
        """

        self._endianness = None
        self._major_version = None
        self._minor_version = None
        self._header = None
        self._close_after = True
        self._interpreter = None

        if not os.path.isfile(file_name):
            raise SarpyIOError('Path {} is not a file'.format(file_name))
        self._file_name = file_name
        self._file_object = open(self._file_name, 'rb')

        check = self._file_object.read(7)
        if check != b'GSATIMG':
            self._file_object.close()
            self._close_after = False
            raise SarpyIOError('file {} is not a GFF file'.format(self._file_name))

        # initialize things
        self._initialize()

    @property
    def file_name(self) -> str:
        """
        str: the file name
        """

        return self._file_name

    @property
    def endianness(self) -> str:
        """
        str: The endian format of the GFF storage. Returns '<' if little-endian
        or '>' if big endian.
        """

        return self._endianness

    @property
    def major_version(self) -> int:
        """
        int: The major GFF version number
        """

        return self._major_version

    @property
    def minor_version(self) -> int:
        """
        int: The minor GFF version number
        """

        return self._minor_version

    @property
    def version(self) -> str:
        """
        str: The GFF version number
        """

        return '{}.{}'.format(self._major_version, self._minor_version)

    @property
    def header(self) -> Union[_GFFHeader_1_6, _GFFHeader_1_8, _GFFHeader_2]:
        """
        The GFF header object.

        Returns
        -------
        _GFFHeader_1_6|_GFFHeader_1_8|_GFFHeader_2
        """

        return self._header

    @property
    def interpreter(self) -> _GFFInterpreter:
        """
        The GFF interpreter object.

        Returns
        -------
        _GFFInterpreter
        """

        return self._interpreter

    def _initialize(self) -> None:
        """
        Initialize the various elements
        """

        self._file_object.seek(7, os.SEEK_SET)
        check = self._file_object.read(1)
        if check == b'\x20':
            # this should be version 1.*, but we will verify
            self._file_object.seek(54, os.SEEK_SET)
            endianness = struct.unpack('H', self._file_object.read(2))[0]  # 0 if little endian
            estr = '<' if endianness == 0 else '>'

            self._file_object.seek(8, os.SEEK_SET)
            self._minor_version, self._major_version = struct.unpack(
                '{}HH'.format(estr), self._file_object.read(4))
        elif check == b'\x00':
            # this should be a version 2.*, but we will verify
            estr = '<'
            self._file_object.seek(16, os.SEEK_SET)
            self._major_version, self._minor_version = struct.unpack(
                '{}HH'.format(estr), self._file_object.read(4))
        else:
            raise ValueError('Got unexpected check byte')

        self._file_object.seek(0, os.SEEK_SET)
        self._endianness = estr
        version = self.version
        if version == '1.6':
            self._header = _GFFHeader_1_6(self._file_object, self.endianness)
            self._interpreter = _GFFInterpreter1(self._header)
        elif version == '1.8':
            self._header = _GFFHeader_1_8(self._file_object, self.endianness)
            self._interpreter = _GFFInterpreter1(self._header)
        elif self.major_version == 2:
            self._header = _GFFHeader_2(self._file_object, self.endianness)
            self._interpreter = _GFFInterpreter2(self._header)
        else:
            raise ValueError('Got unhandled GFF version `{}`'.format(version))

    def get_sicd(self) -> SICDType:
        """
        Gets the sicd structure.

        Returns
        -------
        SICDType
        """

        return self._interpreter.get_sicd()

    def get_data_segment(self) -> DataSegment:
        """
        Gets the data segment.

        Returns
        -------
        DataSegment
        """

        return self._interpreter.get_data_segment()

    def close(self):
        try:
            if self._close_after:
                self._file_object.close()
            if self._interpreter is not None:
                self._interpreter.clean_up()
            self._interpreter = None
        except AttributeError:
            pass

    def __del__(self):
        self.close()


class GFFReader(SICDTypeReader):
    """
    A GFF (Sandia format) reader implementation.

    **Changed in version 1.3.0** for reading changes.
    """

    __slots__ = ('_gff_details', )

    def __init__(self, gff_details: Union[str, GFFDetails]):
        """

        Parameters
        ----------
        gff_details : str|GFFDetails
            file name or GFFDetails object
        """

        if isinstance(gff_details, str):
            gff_details = GFFDetails(gff_details)
        if not isinstance(gff_details, GFFDetails):
            raise TypeError('The input argument for a GFFReader must be a '
                            'filename or GFFDetails object')
        self._gff_details = gff_details

        sicd = gff_details.get_sicd()
        data_segment = gff_details.get_data_segment()
        SICDTypeReader.__init__(self, data_segment, sicd, close_segments=True)
        self._check_sizes()

    @property
    def gff_details(self) -> GFFDetails:
        """
        GFFDetails: The details object.
        """

        return self._gff_details

    @property
    def file_name(self):
        return self.gff_details.file_name

    def close(self) -> None:
        SICDTypeReader.close(self)
        if self._gff_details is not None:
            self._gff_details.close()
        self._gff_details = None

    def __del__(self):
        self.close()


########
# base expected functionality for a module with an implemented Reader

def is_a(file_name: str) -> Optional[GFFReader]:
    """
    Tests whether a given file_name corresponds to a Cosmo Skymed file. Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str|BinaryIO
        the file_name to check

    Returns
    -------
    CSKReader|None
        `CSKReader` instance if Cosmo Skymed file, `None` otherwise
    """

    if is_file_like(file_name):
        return None

    try:
        gff_details = GFFDetails(file_name)
        logger.info('File {} is determined to be a GFF version {} file.'.format(
            file_name, gff_details.version))
        return GFFReader(gff_details)
    except SarpyIOError:
        return None



# MIT License
#
# Copyright (c) 2020 National Geospatial-Intelligence Agency
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


__all__ = ['__version__',
           '__classification__', '__author__', '__url__', '__email__',
           '__title__', '__summary__',
           '__license__', '__copyright__']

from sarpy.__details__ import __classification__, _post_identifier
_version_number = '1.3.60'

__version__ = _version_number + _post_identifier

__author__ = "National Geospatial-Intelligence Agency"
__url__ = "https://github.com/ngageoint/sarpy"
__email__ = "richard.m.naething@nga.mil"


__title__ = "sarpy"
__summary__ = "Python tools for reading, writing, and simple processing of complex SAR data and other " \
              "associated data."


__license__ = "MIT License"
__copyright__ = "2020 {}".format(__author__)



"""
This module provides structures for annotating a given SICD type file for RCS
calculations
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from collections import OrderedDict, defaultdict
import json
from typing import Union, Any, List

import numpy

from sarpy.geometry.geometry_elements import Jsonable, Polygon, MultiPolygon
from sarpy.annotation.base import AnnotationFeature, AnnotationProperties, \
    AnnotationCollection, FileAnnotationCollection

from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.utils import get_im_physical_coords


_RCS_VERSION = "RCS:1.0"
logger = logging.getLogger(__name__)

DEFAULT_NAME_MAPPING = OrderedDict(
    RCS='RCSSFPoly',
    BetaZero='BetaZeroSFPoly',
    GammaZero='GammaZeroSFPoly',
    SigmaZero='SigmaZeroSFPoly')


def _get_polygon_bounds(polygon, data_size):
    """
    Gets the row/column bounds for the polygon and a polygon inclusion mask for
    the defined rectangular pixel grid.

    Parameters
    ----------
    polygon : Polygon
    data_size : Tuple[int, int]

    Returns
    -------
    row_bounds : Tuple[int, int]
        The lower and upper bounds for the rows.
    col_bounds : Tuple[int, int]
        The lower and upper bounds for the columns.
    mask: numpy.ndarray
        The boolean inclusion mask.
    """

    if not isinstance(polygon, Polygon):
        raise TypeError('polygon must be an instance of Polygon, got type {}'.format(type(polygon)))

    bounding_box = polygon.get_bbox()
    if len(bounding_box) != 4:
        raise ValueError('Got unexpected bounding box {}'.format(bounding_box))

    row_min = max(0, int(numpy.floor(bounding_box[0])))
    row_max = min(int(numpy.floor(bounding_box[2])) + 1, data_size[0])

    col_min = max(0, int(numpy.floor(bounding_box[1])))
    col_max = min(int(numpy.floor(bounding_box[3])) + 1, data_size[1])

    row_bounds = (row_min, row_max)
    col_bounds = (col_min, col_max)
    mask = polygon.grid_contained(
        numpy.arange(row_bounds[0], row_bounds[1]),
        numpy.arange(col_bounds[0], col_bounds[1]))
    return row_bounds, col_bounds, mask


def create_rcs_value_collection_for_reader(reader, polygon):
    """
    Given a SICD type reader and a polygon with coordinates in pixel space
    (all sicd footprint assumed applicable), construct the `RCSValueCollection`.

    Parameters
    ----------
    reader : SICDTypeReader
    polygon : Polygon|MultiPolygon

    Returns
    -------
    RCSValueCollection
    """

    def evaluate_sicd(the_sicd):
        # type: (SICDType) -> (bool, bool)
        if the_sicd.Radiometric is None:
            return False, False

        if the_sicd.Radiometric.NoiseLevel is None:
            return True, False
        if the_sicd.Radiometric.NoiseLevel.NoiseLevelType == 'ABSOLUTE':
            return True, True
        else:
            return True, False

    def get_stat_entries():
        def get_empty_dict():
            return dict(total=0.0, total2=0.0, count=0, min=numpy.inf, max=-numpy.inf)

        return defaultdict(
            lambda: dict(value=get_empty_dict(), noise=get_empty_dict()) if has_noise else dict(value=get_empty_dict()))

    def calculate_statistics(array, the_entry):
        # type: (numpy.ndarray, dict) -> None
        the_entry['total'] += numpy.sum(array)
        the_entry['total2'] += numpy.sum(array*array)
        the_entry['count'] += array.size
        the_entry['min'] = min(the_entry['min'], numpy.min(array))
        the_entry['max'] = max(the_entry['max'], numpy.max(array))

    def get_total_rcs(the_stats, the_pol, the_ind, oversamp):
        if has_radiometric:
            the_entry = the_stats['RCS']
            name = 'TotalRCS'
            val = the_entry['value']['total']/oversamp
            noise_val = the_entry['noise']['total']/oversamp if has_noise else None
        else:
            the_entry = the_stats['PixelPower']
            name = 'TotalPixelPower'
            val = the_entry['value']['total']
            noise_val = the_entry['noise']['total'] if has_noise else None

        out = RCSValue(the_pol, name, the_ind, value=RCSStatistics(mean=val))
        if has_noise:
            out.noise = RCSStatistics(mean=noise_val)
        return out

    def get_rcs_value(the_stats, the_pol, name, the_ind):
        def make_stat_entry(vals):
            the_count = vals['count']
            if the_count == 0:
                the_mean = float('NaN')
                the_var = float('NaN')
            else:
                the_mean = float(vals['total']/the_count)
                the_var = vals['total2']/the_count - the_mean*the_mean
                if the_var < 0:
                    the_var = 0  # to avoid floating point errors
            return RCSStatistics(
                mean=the_mean, std=float(numpy.sqrt(the_var)), min=float(vals['min']), max=float(vals['max']))

        the_entry = the_stats[name]
        noise_value = the_entry.get('noise', None)

        out = RCSValue(the_pol, name, the_ind)
        out.value = make_stat_entry(the_entry['value'])
        if noise_value is not None:
            out.noise = make_stat_entry(noise_value)
        return out

    # verify that all footprint are identical
    data_sizes = reader.get_data_size_as_tuple()
    if len(data_sizes) > 1:
        for entry in data_sizes[1:]:
            if entry != data_sizes[0]:
                raise ValueError('Each image index must have identical size')
    data_size = data_sizes[0]

    if isinstance(polygon, Polygon):
        polygons = [polygon, ]
    elif isinstance(polygon, MultiPolygon):
        polygons = polygon.polygons
    else:
        raise TypeError('polygon must be a Polygon or MultiPolygon, got type {}'.format(type(polygon)))

    sicds = reader.get_sicds_as_tuple()
    radiometric_signature = None
    for sicd in sicds:
        if radiometric_signature is None:
            radiometric_signature = evaluate_sicd(sicd)
        elif radiometric_signature != evaluate_sicd(sicd):
            raise ValueError('All sicds in the reader must have compatible Radiometric definition')
    has_radiometric, has_noise = radiometric_signature

    # construct the statistics values - first/second moments and max/min
    stat_values = [get_stat_entries() for _ in sicds]

    for polygon in polygons:
        row_bounds, col_bounds, mask = _get_polygon_bounds(polygon, data_size)
        if not numpy.any(mask):
            continue

        for i, sicd in enumerate(sicds):
            current_stat_entries = stat_values[i]

            # define the pixel power array for the given polygon and image index
            data = reader[row_bounds[0]:row_bounds[1], col_bounds[0]:col_bounds[1], i][mask]
            data = data.real * data.real + data.imag * data.imag  # get pixel power

            # define the pixel power statistics
            calculate_statistics(data, current_stat_entries['PixelPower']['value'])

            if has_radiometric:
                noise_poly = sicd.Radiometric.NoiseLevel.NoisePoly if has_noise else None
                # construct the physical coordinate arrays
                row_array = numpy.arange(row_bounds[0], row_bounds[1], 1, dtype=numpy.int32)
                x_array = get_im_physical_coords(row_array, sicd.Grid, sicd.ImageData, 'Row')
                col_array = numpy.arange(col_bounds[0], col_bounds[1], 1, dtype=numpy.int32)
                y_array = get_im_physical_coords(col_array, sicd.Grid, sicd.ImageData, 'Col')
                yarr, xarr = numpy.meshgrid(y_array, x_array)
                xarr = xarr[mask]
                yarr = yarr[mask]

                noise_power = numpy.exp(numpy.log(10)*0.1*noise_poly(xarr, yarr)) if has_noise else None
                if has_noise:
                    # add the noise statistics for the pixel power
                    calculate_statistics(noise_power, current_stat_entries['PixelPower']['noise'])

                for units_name, rcs_poly_name in DEFAULT_NAME_MAPPING.items():
                    the_poly = getattr(sicd.Radiometric, rcs_poly_name)
                    sf_data = the_poly(xarr, yarr)
                    calculate_statistics(sf_data*data, current_stat_entries[units_name]['value'])
                    if has_noise:
                        calculate_statistics(sf_data*noise_power, current_stat_entries[units_name]['noise'])

    # convert this collection of raw data to the RCSStatistics collection
    rcs_values = RCSValueCollection()
    for i, sicd in enumerate(sicds):
        polarization = sicd.get_processed_polarization()
        oversample = sicd.Grid.Row.get_oversample_rate()*sicd.Grid.Col.get_oversample_rate()
        raw_stats = stat_values[i]
        # create the total rcs/power entry
        rcs_values.insert_new_element(get_total_rcs(raw_stats, polarization, i, oversample))
        rcs_values.insert_new_element(get_rcs_value(raw_stats, polarization, 'PixelPower', i))
        if has_radiometric:
            for the_units in DEFAULT_NAME_MAPPING.keys():
                rcs_values.insert_new_element(get_rcs_value(raw_stats, polarization, the_units, i))
    return rcs_values


class RCSStatistics(Jsonable):
    __slots__ = ('mean', 'std', 'max', 'min')
    _type = 'RCSStatistics'

    def __init__(self, mean=None, std=None, max=None, min=None):
        """

        Parameters
        ----------
        mean : None|float
            All values are assumed the be stored here in units of power
        std : None|float
            All values are assumed the be stored here in units of power
        max : None|float
            All values are assumed the be stored here in units of power
        min : None|float
            All values are assumed the be stored here in units of power
        """

        if mean is not None:
            mean = float(mean)
        if std is not None:
            std = float(std)
        if max is not None:
            max = float(max)
        if min is not None:
            min = float(min)

        self.mean = mean  # type: Union[None, float]
        self.std = std  # type: Union[None, float]
        self.max = max  # type: Union[None, float]
        self.min = min  # type: Union[None, float]

    @classmethod
    def from_dict(cls, the_json):
        typ = the_json['type']
        if typ != cls._type:
            raise ValueError('RCSStatistics cannot be constructed from {}'.format(the_json))
        return cls(
            mean=the_json.get('mean', None),
            std=the_json.get('std', None),
            max=the_json.get('max', None),
            min=the_json.get('min', None))

    def to_dict(self, parent_dict=None):
        if parent_dict is None:
            parent_dict = OrderedDict()
        parent_dict['type'] = self.type
        for attr in self.__slots__:
            parent_dict[attr] = getattr(self, attr)
        return parent_dict

    def get_field_list(self):
        if self.mean is None:
            return '', '', '', '', ''
        else:
            mean_db_str = '' if self.mean <= 0 else '{0:0.5G}'.format(10*numpy.log10(self.mean))
            return (
                mean_db_str,
                '{0:0.5G}'.format(self.mean),
                '{0:0.5G}'.format(self.std) if self.std is not None else '',
                '{0:0.5G}'.format(self.min) if self.min is not None else '',
                '{0:0.5G}'.format(self.max) if self.max is not None else '',
            )


class RCSValue(Jsonable):
    """
    The collection of RCSStatistics elements.
    """

    __slots__ = ('polarization', 'units', '_index', '_value', '_noise')
    _type = 'RCSValue'

    def __init__(self, polarization, units, index, value=None, noise=None):
        """

        Parameters
        ----------
        polarization : str
        units: str
        index : int
        value : None|RCSStatistics
        noise : None|RCSStatistics
        """
        self._value = None
        self._noise = None
        self._index = None
        self.polarization = polarization
        self.units = units
        self.index = index
        self.value = value
        self.noise = noise

    @property
    def value(self):
        """
        None|RCSStatistics: The value
        """

        return self._value

    @value.setter
    def value(self, val):
        if isinstance(val, dict):
            val = RCSStatistics.from_dict(val)
        if not (val is None or isinstance(val, RCSStatistics)):
            raise TypeError('Got incompatible input for value')
        self._value = val

    @property
    def index(self):
        """
        int: The image index to which this applies
        """

        return self._index

    @index.setter
    def index(self, value):
        if value is None:
            value = 0
        self._index = int(value)

    @property
    def noise(self):
        """
        None|RCSStatistics: The noise
        """

        return self._noise

    @noise.setter
    def noise(self, val):
        if isinstance(val, dict):
            val = RCSStatistics.from_dict(val)
        if not (val is None or isinstance(val, RCSStatistics)):
            raise TypeError('Got incompatible input for noise')
        self._noise = val

    @classmethod
    def from_dict(cls, the_json):  # type: (dict) -> RCSValue
        typ = the_json['type']
        if typ != cls._type:
            raise ValueError('RCSValue cannot be constructed from {}'.format(the_json))
        return cls(
            the_json.get('polarization', None),
            the_json.get('units', None),
            the_json.get('index', None),
            value=the_json.get('value', None),
            noise=the_json.get('noise', None))

    def to_dict(self, parent_dict=None):
        if parent_dict is None:
            parent_dict = OrderedDict()
        parent_dict['type'] = self.type
        parent_dict['polarization'] = self.polarization
        parent_dict['units'] = self.units
        parent_dict['index'] = self.index
        if self.value is not None:
            parent_dict['value'] = self.value.to_dict()
        if self.noise is not None:
            parent_dict['noise'] = self.noise.to_dict()
        return parent_dict


class RCSValueCollection(Jsonable):
    """
    A specific type for the AnnotationProperties.parameters
    """

    __slots__ = ('_pixel_count', '_elements')
    _type = 'RCSValueCollection'

    def __init__(self, pixel_count=None, elements=None):
        """

        Parameters
        ----------
        pixel_count : None|int
        elements : None|List[RCSValue|dict]
        """

        self._pixel_count = None
        self._elements = []

        self.pixel_count = pixel_count
        self.elements = elements

    def __len__(self):
        return len(self._elements)

    def __getitem__(self, item):
        # type: (Union[int, str]) -> Union[None, RCSValue]
        return self._elements[item]

    @property
    def pixel_count(self):
        # type: () -> Union[None, int]
        """
        None|int: The number of integer pixel grid elements contained in the interior
        of the associated geometry element.
        """

        return self._pixel_count

    @pixel_count.setter
    def pixel_count(self, value):
        if value is None:
            self._pixel_count = None
            return
        if not isinstance(value, int):
            value = int(value)
        self._pixel_count = value

    @property
    def elements(self):
        # type: () -> Union[None, List[RCSValue]]
        """
        List[RCSValue]: The RCSValue elements.
        """

        return self._elements

    @elements.setter
    def elements(self, elements):
        if elements is None:
            self._elements = []
            return

        if not isinstance(elements, list):
            raise TypeError('elements must be a list of RCSValue elements')
        self._elements = []
        for element in elements:
            self.insert_new_element(element)

    def insert_new_element(self, element):
        """
        Inserts an element at the end of the elements list.

        Parameters
        ----------
        element : RCSValue
        """

        if isinstance(element, dict):
            element = RCSValue.from_dict(element)
        if not isinstance(element, RCSValue):
            raise TypeError('element must be an RCSValue instance')
        self._elements.append(element)

    @classmethod
    def from_dict(cls, the_json):
        # type: (dict) -> RCSValueCollection

        typ = the_json['type']
        if typ != cls._type:
            raise ValueError('RCSValueCollection cannot be constructed from {}'.format(the_json))
        return cls(
            pixel_count=the_json.get('pixel_count', None), elements=the_json.get('elements', None))

    def to_dict(self, parent_dict=None):
        if parent_dict is None:
            parent_dict = OrderedDict()
        parent_dict['type'] = self.type
        parent_dict['pixel_count'] = self.pixel_count
        if len(self._elements) > 0:
            parent_dict['elements'] = [entry.to_dict() for entry in self._elements]
        return parent_dict


class RCSProperties(AnnotationProperties):
    _type = 'RCSProperties'

    @property
    def parameters(self):
        """
        RCSValueCollection: The parameters
        """

        return self._parameters

    @parameters.setter
    def parameters(self, value):
        if value is None:
            self._parameters = RCSValueCollection()
            return

        if isinstance(value, dict):
            value = RCSValueCollection.from_dict(value)
        if not isinstance(value, RCSValueCollection):
            raise TypeError('Got unexpected type for parameters')
        self._parameters = value


class RCSFeature(AnnotationFeature):
    """
    A specific extension of the Feature class which has the properties attribute
    populated with RCSValueCollection instance.
    """
    _allowed_geometries = (Polygon, MultiPolygon)

    @property
    def properties(self):
        # type: () -> RCSProperties
        """
        The properties.

        Returns
        -------
        RCSProperties
        """

        return self._properties

    @properties.setter
    def properties(self, properties):
        if properties is None:
            self._properties = RCSProperties()
        elif isinstance(properties, RCSProperties):
            self._properties = properties
        elif isinstance(properties, dict):
            self._properties = RCSProperties.from_dict(properties)
        else:
            raise TypeError('properties must be an RCSProperties')

    def set_rcs_parameters_from_reader(self, reader):
        """
        Given a SICD type reader construct the `RCSValueCollection` and set that
        as the properties.parameters value.

        Parameters
        ----------
        reader : SICDTypeReader
        """

        if self.geometry is None or self.geometry_count == 0:
            self.properties.parameters = None
        else:
            # noinspection PyTypeChecker
            self.properties.parameters = create_rcs_value_collection_for_reader(
                reader, self.geometry)


class RCSCollection(AnnotationCollection):
    """
    A specific extension of the AnnotationCollection class which has that the
    features are RCSFeature instances.
    """

    @property
    def features(self):
        """
        The features list.

        Returns
        -------
        List[RCSFeature]
        """

        return self._features

    @features.setter
    def features(self, features):
        if features is None:
            self._features = None
            self._feature_dict = None
            return

        if not isinstance(features, list):
            raise TypeError('features must be a list of RCSFeatures. Got {}'.format(type(features)))

        for entry in features:
            self.add_feature(entry)

    def add_feature(self, feature):
        """
        Add an annotation.

        Parameters
        ----------
        feature : RCSFeature|dict
        """

        if isinstance(feature, dict):
            feature = RCSFeature.from_dict(feature)

        if not isinstance(feature, RCSFeature):
            raise TypeError('This requires an RCSFeature instance, got {}'.format(type(feature)))

        if self._features is None:
            self._feature_dict = {feature.uid: 0}
            self._features = [feature, ]
        else:
            self._feature_dict[feature.uid] = len(self._features)
            self._features.append(feature)

    def __getitem__(self, item):
        # type: (Any) -> Union[RCSFeature, List[RCSFeature]]
        if self._features is None:
            raise StopIteration

        if isinstance(item, str):
            index = self._feature_dict[item]
            return self._features[index]
        return self._features[item]


###########
# serialized file object

class FileRCSCollection(FileAnnotationCollection):
    """
    An collection of RCS statistics elements.
    """
    _type = 'FileRCSCollection'

    def __init__(self, version=None, annotations=None, image_file_name=None,
                 image_id=None, core_name=None):
        if version is None:
            version = _RCS_VERSION

        FileAnnotationCollection.__init__(
            self, version=version, annotations=annotations, image_file_name=image_file_name,
            image_id=image_id, core_name=core_name)

    @property
    def annotations(self):
        """
        The annotations.

        Returns
        -------
        RCSCollection
        """

        return self._annotations

    @annotations.setter
    def annotations(self, annotations):
        # type: (Union[None, RCSCollection, dict]) -> None
        if annotations is None:
            self._annotations = None
            return

        if isinstance(annotations, RCSCollection):
            self._annotations = annotations
        elif isinstance(annotations, dict):
            self._annotations = RCSCollection.from_dict(annotations)
        else:
            raise TypeError(
                'annotations must be an RCSCollection. Got type {}'.format(type(annotations)))

    def add_annotation(self, annotation):
        """
        Add an annotation.

        Parameters
        ----------
        annotation : RCSFeature
            The prospective annotation.
        """

        if isinstance(annotation, dict):
            annotation = RCSFeature.from_dict(annotation)
        if not isinstance(annotation, RCSFeature):
            raise TypeError('This requires an RCSFeature instance. Got {}'.format(type(annotation)))

        if self._annotations is None:
            self._annotations = RCSCollection()

        self._annotations.add_feature(annotation)

    def delete_annotation(self, annotation_id):
        """
        Deletes the annotation associated with the given id.

        Parameters
        ----------
        annotation_id : str
        """

        del self._annotations[annotation_id]

    @classmethod
    def from_file(cls, file_name):
        """
        Read from (json) file.

        Parameters
        ----------
        file_name : str

        Returns
        -------
        FileRCSCollection
        """

        with open(file_name, 'r') as fi:
            the_dict = json.load(fi)
        return cls.from_dict(the_dict)

    @classmethod
    def from_dict(cls, the_dict):
        """
        Define from a dictionary representation.

        Parameters
        ----------
        the_dict : dict

        Returns
        -------
        FileRCSCollection
        """

        if not isinstance(the_dict, dict):
            raise TypeError('This requires a dict. Got type {}'.format(type(the_dict)))

        typ = the_dict.get('type', 'NONE')
        if typ != cls._type:
            raise ValueError('FileRCSCollection cannot be constructed from the input dictionary')

        return cls(
            version=the_dict.get('version', 'UNKNOWN'),
            annotations=the_dict.get('annotations', None),
            image_file_name=the_dict.get('image_file_name', None),
            image_id=the_dict.get('image_id', None),
            core_name=the_dict.get('core_name', None))




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class EPHEM(TREElement):
    def __init__(self, value):
        super(EPHEM, self).__init__()
        self.add_field('EPHEM_X', 's', 12, value)
        self.add_field('EPHEM_Y', 's', 12, value)
        self.add_field('EPHEM_Z', 's', 12, value)


class CSEPHAType(TREElement):
    def __init__(self, value):
        super(CSEPHAType, self).__init__()
        self.add_field('EPHEM_FLAG', 's', 12, value)
        self.add_field('DT_EPHEM', 's', 5, value)
        self.add_field('DATE_EPHEM', 's', 8, value)
        self.add_field('T0_EPHEM', 's', 13, value)
        self.add_field('NUM_EPHEM', 'd', 3, value)
        self.add_loop('EPHEMs', self.NUM_EPHEM, EPHEM, value)


class CSEPHA(TREExtension):
    _tag_value = 'CSEPHA'
    _data_type = CSEPHAType



"""
The Compensated Received Signal Data 1.0 definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Michael Stewart, Valkyrie")

from xml.etree import ElementTree
from collections import OrderedDict
from typing import Union
import numpy

from sarpy.io.xml.base import Serializable, find_children, parse_xml_from_file, \
    parse_xml_from_string
from sarpy.io.xml.descriptors import SerializableDescriptor, IntegerDescriptor, \
    StringDescriptor

from .base import DEFAULT_STRICT
from .CollectionID import CollectionIDType
from .Global import GlobalType
from sarpy.io.phase_history.cphd1_elements.SceneCoordinates import SceneCoordinatesType
from .Data import DataType
from .Channel import ChannelType
from .PVP import PVPType
from sarpy.io.phase_history.cphd1_elements.SupportArray import SupportArrayType
from sarpy.io.phase_history.cphd1_elements.Dwell import DwellType
from .ReferenceGeometry import ReferenceGeometryType
from .Antenna import AntennaType
from .ErrorParameters import ErrorParametersType
from sarpy.io.phase_history.cphd1_elements.ProductInfo import ProductInfoType
from sarpy.io.phase_history.cphd1_elements.GeoInfo import GeoInfoType
from sarpy.io.complex.sicd_elements.MatchInfo import MatchInfoType

from sarpy.io.phase_history.cphd_schema import get_urn_details, WRITABLE_VERSIONS, \
    get_namespace, get_default_tuple


#########
# Module variables

#########
# Module variables
_CRSD_SPEC_DETAILS = {
    key: {'namespace': get_namespace(key), 'details': get_urn_details(key)}
    for key in WRITABLE_VERSIONS}

_CRSD_DEFAULT_TUPLE = get_default_tuple()
_CRSD_DEFAULT_VERSION = '{}.{}.{}'.format(*_CRSD_DEFAULT_TUPLE)
CRSD_SECTION_TERMINATOR = b'\f\n'


#########
# CRSD header object

def _parse_crsd_header_field(line):
    """
    Parse the CRSD header field, or return `None` as a termination signal.

    Parameters
    ----------
    line : bytes

    Returns
    -------
    None|(str, str)
    """

    if line.startswith(CRSD_SECTION_TERMINATOR):
        return None
    parts = line.split(b' := ')
    if len(parts) != 2:
        raise ValueError('Cannot extract CRSD header value from line {}'.format(line))
    fld = parts[0].strip().decode('utf-8')
    val = parts[1].strip().decode('utf-8')
    return fld, val


class CRSDHeaderBase(object):
    _fields = ()
    _required = ()

    def __init__(self, **kwargs):
        # abstract class
        pass

    @classmethod
    def from_file_object(cls, fi):
        """
        Extract the CRSD header object from a file opened in byte mode.
        This file object is assumed to be at the correct location for the
        CRSD header.

        Parameters
        ----------
        fi
            The open file object, which will be progressively read.

        Returns
        -------
        CRSDHeaderBase
        """

        the_dict = {}
        while True:
            line = fi.readline()
            res = _parse_crsd_header_field(line)
            if res is None:
                break
            else:
                fld, val = res
            if fld not in cls._fields:
                raise ValueError('Cannot extract CRSD header value from line {}'.format(line))
            the_dict[fld] = val
        return cls(**the_dict)


class CRSDHeader(CRSDHeaderBase):
    _fields = (
        'XML_BLOCK_SIZE', 'XML_BLOCK_BYTE_OFFSET', 'SUPPORT_BLOCK_SIZE', 'SUPPORT_BLOCK_BYTE_OFFSET',
        'PVP_BLOCK_SIZE', 'PVP_BLOCK_BYTE_OFFSET', 'SIGNAL_BLOCK_SIZE', 'SIGNAL_BLOCK_BYTE_OFFSET',
        'CLASSIFICATION', 'RELEASE_INFO')
    _required = (
        'XML_BLOCK_SIZE', 'XML_BLOCK_BYTE_OFFSET', 'PVP_BLOCK_SIZE', 'PVP_BLOCK_BYTE_OFFSET',
        'SIGNAL_BLOCK_SIZE', 'SIGNAL_BLOCK_BYTE_OFFSET', 'CLASSIFICATION', 'RELEASE_INFO')
    # descriptor
    XML_BLOCK_SIZE = IntegerDescriptor(
        'XML_BLOCK_SIZE', _required, strict=True,
        docstring='Size of the XML instance that describes the product in bytes. '
                  'Size does NOT include the 2 bytes of the section terminator.')  # type: int
    XML_BLOCK_BYTE_OFFSET = IntegerDescriptor(
        'XML_BLOCK_BYTE_OFFSET', _required, strict=True,
        docstring='Offset to the first byte of the XML block in bytes.')  # type: int
    SUPPORT_BLOCK_SIZE = IntegerDescriptor(
        'SUPPORT_BLOCK_SIZE', _required, strict=True,
        docstring='Size of the Support block in bytes. Note - If the Support block is omitted, this '
                  'is not included.')  # type: int
    SUPPORT_BLOCK_BYTE_OFFSET = IntegerDescriptor(
        'SUPPORT_BLOCK_BYTE_OFFSET', _required, strict=True,
        docstring='Offset to the first byte of the Support block in bytes. Note - If the Support '
                  'block is omitted, this is not included.')  # type: int
    PVP_BLOCK_SIZE = IntegerDescriptor(
        'PVP_BLOCK_SIZE', _required, strict=True,
        docstring='Size of the PVP block in bytes.')  # type: int
    PVP_BLOCK_BYTE_OFFSET = IntegerDescriptor(
        'PVP_BLOCK_BYTE_OFFSET', _required, strict=True,
        docstring='Offset to the first byte of the PVP block in bytes.')  # type: int
    SIGNAL_BLOCK_SIZE = IntegerDescriptor(
        'SIGNAL_BLOCK_SIZE', _required, strict=True,
        docstring='Size of the Signal block in bytes.')  # type: int
    SIGNAL_BLOCK_BYTE_OFFSET = IntegerDescriptor(
        'SIGNAL_BLOCK_BYTE_OFFSET', _required, strict=True,
        docstring='Offset to the first byte of the Signal block in bytes.')  # type: int
    CLASSIFICATION = StringDescriptor(
        'CLASSIFICATION', _required, strict=True, default_value='UNCLASSIFIED',
        docstring='Product classification information that is human-readable.')  # type: str
    RELEASE_INFO = StringDescriptor(
        'RELEASE_INFO', _required, strict=True, default_value='UNRESTRICTED',
        docstring='Product release information that is human-readable.')  # type: str

    def __init__(self, XML_BLOCK_SIZE=None, XML_BLOCK_BYTE_OFFSET=None,
                 SUPPORT_BLOCK_SIZE=None, SUPPORT_BLOCK_BYTE_OFFSET=None,
                 PVP_BLOCK_SIZE=None, PVP_BLOCK_BYTE_OFFSET=None,
                 SIGNAL_BLOCK_SIZE=None, SIGNAL_BLOCK_BYTE_OFFSET=None,
                 CLASSIFICATION='UNCLASSIFIED', RELEASE_INFO='UNRESTRICTED',
                 use_version=None):
        self.XML_BLOCK_SIZE = XML_BLOCK_SIZE
        self.XML_BLOCK_BYTE_OFFSET = XML_BLOCK_BYTE_OFFSET
        self.SUPPORT_BLOCK_SIZE = SUPPORT_BLOCK_SIZE
        self.SUPPORT_BLOCK_BYTE_OFFSET = SUPPORT_BLOCK_BYTE_OFFSET
        self.PVP_BLOCK_SIZE = PVP_BLOCK_SIZE
        self.PVP_BLOCK_BYTE_OFFSET = PVP_BLOCK_BYTE_OFFSET
        self.SIGNAL_BLOCK_SIZE = SIGNAL_BLOCK_SIZE
        self.SIGNAL_BLOCK_BYTE_OFFSET = SIGNAL_BLOCK_BYTE_OFFSET
        self.CLASSIFICATION = CLASSIFICATION
        self.RELEASE_INFO = RELEASE_INFO
        self._use_version = _CRSD_DEFAULT_VERSION if use_version is None else use_version
        super(CRSDHeader, self).__init__()

    @property
    def use_version(self) -> str:
        return self._use_version

    def to_string(self):
        """
        Forms a CRSD file header string (not including the section terminator) from populated attributes.
        """
        return ('CRSD/{}\n'.format(self.use_version)
                + ''.join(["{} := {}\n".format(f, getattr(self, f))
                           for f in self._fields if getattr(self, f) is not None]))


class CRSDType(Serializable):
    """
    The Compensated Received Signal Data definition.
    """

    _fields = (
        'CollectionID', 'Global', 'SceneCoordinates', 'Data', 'Channel', 'PVP',
        'SupportArray', 'Dwell', 'ReferenceGeometry', 'Antenna',
        'ErrorParameters', 'ProductInfo', 'GeoInfo', 'MatchInfo')
    _required = (
        'CollectionID', 'Global', 'Data', 'Channel', 'PVP', 'ReferenceGeometry')
    _collections_tags = {'GeoInfo': {'array': 'False', 'child_tag': 'GeoInfo'}}
    # descriptors
    CollectionID = SerializableDescriptor(
        'CollectionID', CollectionIDType, _required, strict=DEFAULT_STRICT,
        docstring='General information about the collection.')  # type: CollectionIDType
    Global = SerializableDescriptor(
        'Global', GlobalType, _required, strict=DEFAULT_STRICT,
        docstring='Global parameters that apply to metadata components and CRSD '
                  'signal arrays.')  # type: GlobalType
    SceneCoordinates = SerializableDescriptor(
        'SceneCoordinates', SceneCoordinatesType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that define geographic coordinates of the imaged '
                  'scene.')  # type: Union[None, SceneCoordinatesType]
    Data = SerializableDescriptor(
        'Data', DataType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe binary data components contained in '
                  'the product.')  # type: DataType
    Channel = SerializableDescriptor(
        'Channel', ChannelType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the data channels contained in the '
                  'product.')  # type: ChannelType
    PVP = SerializableDescriptor(
        'PVP', PVPType, _required, strict=DEFAULT_STRICT,
        docstring='Structure used to specify the Per Vector Parameters provided for '
                  'each channel of a given product.')  # type: PVPType
    SupportArray = SerializableDescriptor(
        'SupportArray', SupportArrayType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the binary support array(s) content and '
                  'grid coordinates.')  # type: Union[None, SupportArrayType]
    Dwell = SerializableDescriptor(
        'Dwell', DwellType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that specify the SAR dwell time supported by the signal '
                  'arrays contained in the CRSD product.')  # type: Union[None, DwellType]
    ReferenceGeometry = SerializableDescriptor(
        'ReferenceGeometry', ReferenceGeometryType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the collection geometry for the reference '
                  'vector of the reference channel.')  # type: ReferenceGeometryType
    Antenna = SerializableDescriptor(
        'Antenna', AntennaType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the antennas antennas used '
                  'to collect the signal array(s).')  # type: Union[None, AntennaType]
    ErrorParameters = SerializableDescriptor(
        'ErrorParameters', ErrorParametersType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the statistics of errors in measured or estimated parameters '
                  'that describe the collection.')  # type: Union[None, ErrorParametersType]
    ProductInfo = SerializableDescriptor(
        'ProductInfo', ProductInfoType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that provide general information about the CRSD product '
                  'and/or the derived products that may be created '
                  'from it.')  # type: Union[None, ProductInfoType]
    MatchInfo = SerializableDescriptor(
        'MatchInfo', MatchInfoType, _required, strict=DEFAULT_STRICT,
        docstring='Information about other collections that are matched to the collection from which '
                  'this CRSD product was generated.')  # type: Union[None, MatchInfoType]

    def __init__(self, CollectionID=None, Global=None, SceneCoordinates=None, Data=None,
                 Channel=None, PVP=None, SupportArray=None, Dwell=None, ReferenceGeometry=None,
                 Antenna=None, ErrorParameters=None, ProductInfo=None,
                 GeoInfo=None, MatchInfo=None, **kwargs):
        """

        Parameters
        ----------
        CollectionID : CollectionIDType
        Global : GlobalType
        SceneCoordinates : None|SceneCoordinatesType
        Data : DataType
        Channel : ChannelType
        PVP : PVPType
        SupportArray : None|SupportArrayType
        Dwell : None|DwellType
        ReferenceGeometry : ReferenceGeometryType
        Antenna : None|AntennaType
        ErrorParameters : None|ErrorParametersType
        ProductInfo : None|ProductInfoType
        GeoInfo : None|List[GeoInfoType]|GeoInfoType
        MatchInfo : None|MatchInfoType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CollectionID = CollectionID
        self.Global = Global
        self.SceneCoordinates = SceneCoordinates
        self.Data = Data
        self.Channel = Channel
        self.PVP = PVP
        self.SupportArray = SupportArray
        self.Dwell = Dwell
        self.ReferenceGeometry = ReferenceGeometry
        self.Antenna = Antenna
        self.ErrorParameters = ErrorParameters
        self.ProductInfo = ProductInfo
        self.MatchInfo = MatchInfo

        self._GeoInfo = []
        if GeoInfo is None:
            pass
        elif isinstance(GeoInfo, GeoInfoType):
            self.addGeoInfo(GeoInfo)
        elif isinstance(GeoInfo, (list, tuple)):
            for el in GeoInfo:
                self.addGeoInfo(el)
        else:
            raise ValueError('GeoInfo got unexpected type {}'.format(type(GeoInfo)))

        super(CRSDType, self).__init__(**kwargs)

    @property
    def GeoInfo(self):
        """
        List[GeoInfoType]: Parameters that describe a geographic feature.
        """

        return self._GeoInfo

    def getGeoInfo(self, key):
        """
        Get GeoInfo(s) with name attribute == `key`.

        Parameters
        ----------
        key : str

        Returns
        -------
        List[GeoInfoType]
        """

        return [entry for entry in self._GeoInfo if entry.name == key]

    def addGeoInfo(self, value):
        """
        Add the given GeoInfo to the GeoInfo list.

        Parameters
        ----------
        value : GeoInfoType

        Returns
        -------
        None
        """

        if isinstance(value, ElementTree.Element):
            gi_key = self._child_xml_ns_key.get('GeoInfo', self._xml_ns_key)
            value = GeoInfoType.from_node(value, self._xml_ns, ns_key=gi_key)
        elif isinstance(value, dict):
            value = GeoInfoType.from_dict(value)

        if isinstance(value, GeoInfoType):
            self._GeoInfo.append(value)
        else:
            raise TypeError('Trying to set GeoInfo element with unexpected type {}'.format(type(value)))

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if kwargs is None:
            kwargs = OrderedDict()
        gi_key = cls._child_xml_ns_key.get('GeoInfo', ns_key)
        kwargs['GeoInfo'] = find_children(node, 'GeoInfo', xml_ns, gi_key)
        return super(CRSDType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        node = super(CRSDType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity,
            strict=strict, exclude=exclude+('GeoInfo', ))
        # slap on the GeoInfo children
        if self._GeoInfo is not None and len(self._GeoInfo) > 0:
            for entry in self._GeoInfo:
                entry.to_node(doc, 'GeoInfo', ns_key=ns_key, parent=node, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = super(CRSDType, self).to_dict(
            check_validity=check_validity, strict=strict, exclude=exclude+('GeoInfo', ))
        # slap on the GeoInfo children
        if len(self.GeoInfo) > 0:
            out['GeoInfo'] = [entry.to_dict(
                check_validity=check_validity, strict=strict) for entry in self._GeoInfo]
        return out

    def to_xml_bytes(self, urn=None, tag='CRSD', check_validity=False, strict=DEFAULT_STRICT):
        if urn is None:
            urn = get_namespace(_CRSD_DEFAULT_VERSION)
        return super(CRSDType, self).to_xml_bytes(
            urn=urn, tag=tag, check_validity=check_validity, strict=strict)

    def to_xml_string(self, urn=None, tag='CRSD', check_validity=False, strict=DEFAULT_STRICT):
        return self.to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict).decode('utf-8')

    def make_file_header(self, xml_offset=1024, use_version=None):
        """
        Forms a CRSD file header consistent with the information in the Data and CollectionID nodes.

        Parameters
        ----------
        xml_offset : int, optional
            Offset in bytes to the first byte of the XML block. If the provided value
            is not large enough to account for the length of the file header
            string, a larger value is chosen.
        use_version : None|str
            What version to use?

        Returns
        -------
        header : CRSDHeader
        """

        kwargs = OrderedDict()
        kwargs['use_version'] = _CRSD_DEFAULT_VERSION if use_version is None else use_version

        def _align(val):
            align_to = 64
            return int(numpy.ceil(float(val)/align_to)*align_to)

        kwargs['XML_BLOCK_SIZE'] = len(self.to_xml_string())
        kwargs['XML_BLOCK_BYTE_OFFSET'] = xml_offset
        block_end = kwargs['XML_BLOCK_BYTE_OFFSET'] + kwargs['XML_BLOCK_SIZE'] + len(CRSD_SECTION_TERMINATOR)

        if self.Data.NumSupportArrays > 0:
            kwargs['SUPPORT_BLOCK_SIZE'] = self.Data.calculate_support_block_size()
            kwargs['SUPPORT_BLOCK_BYTE_OFFSET'] = _align(block_end)
            block_end = kwargs['SUPPORT_BLOCK_BYTE_OFFSET'] + kwargs['SUPPORT_BLOCK_SIZE']

        kwargs['PVP_BLOCK_SIZE'] = self.Data.calculate_pvp_block_size()
        kwargs['PVP_BLOCK_BYTE_OFFSET'] = _align(block_end)
        block_end = kwargs['PVP_BLOCK_BYTE_OFFSET'] + kwargs['PVP_BLOCK_SIZE']

        kwargs['SIGNAL_BLOCK_SIZE'] = self.Data.calculate_signal_block_size()
        kwargs['SIGNAL_BLOCK_BYTE_OFFSET'] = _align(block_end)
        kwargs['CLASSIFICATION'] = self.CollectionID.Classification
        kwargs['RELEASE_INFO'] = self.CollectionID.ReleaseInfo

        header = CRSDHeader(**kwargs)
        header_str = header.to_string()
        min_xml_offset = len(header_str) + len(CRSD_SECTION_TERMINATOR)
        if kwargs['XML_BLOCK_BYTE_OFFSET'] < min_xml_offset:
            header = self.make_file_header(xml_offset=_align(min_xml_offset + 32), use_version=use_version)
        return header

    def get_pvp_dtype(self):
        """
        Gets the dtype for the corresponding PVP structured array. Note that they
        must all have homogeneous dtype.

        Returns
        -------
        numpy.dtype
            This will be a compound dtype for a structured array.
        """

        if self.PVP is None:
            raise ValueError('No PVP defined.')
        return self.PVP.get_vector_dtype()

    @classmethod
    def from_xml_file(cls, file_path):
        """
        Construct the crsd object from a stand-alone xml file path.

        Parameters
        ----------
        file_path : str

        Returns
        -------
        CRSDType
        """

        root_node, xml_ns = parse_xml_from_file(file_path)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)

    @classmethod
    def from_xml_string(cls, xml_string):
        """
        Construct the crsd object from an xml string.

        Parameters
        ----------
        xml_string : str|bytes

        Returns
        -------
        CRSDType
        """

        root_node, xml_ns = parse_xml_from_string(xml_string)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)

    def version_required(self):
        """
        What CPHD version is required for valid support?

        Returns
        -------
        tuple
        """

        required = (1, 0, 0)
        for fld in self._fields:
            val = getattr(self, fld)
            if val is not None and hasattr(val, 'version_required'):
                required = max(required, val.version_required())
        return required
